@@ -29,30 +29,33 @@
 #include <sys/stat.h>
 #include <sys/inotify.h>
 #include <fcntl.h>
+#include <dirent.h>
 
 #include <libudev.h>
 #include <alsa/asoundlib.h>
 
-#include <spa/support/log.h>
 #include <spa/utils/type.h>
 #include <spa/utils/keys.h>
 #include <spa/utils/names.h>
 #include <spa/utils/result.h>
+#include <spa/utils/string.h>
 #include <spa/support/loop.h>
 #include <spa/support/plugin.h>
 #include <spa/monitor/device.h>
 #include <spa/monitor/utils.h>
 
-#define NAME  "alsa-udev"
+#include "alsa.h"
 
 #define MAX_DEVICES	64
 
 #define ACTION_ADD	0
 #define ACTION_REMOVE	1
+#define ACTION_DISABLE	2
 
 struct device {
 	uint32_t id;
 	struct udev_device *dev;
+	unsigned int unavailable:1;
 	unsigned int accessible:1;
 	unsigned int ignored:1;
 	unsigned int emitted:1;
@@ -64,6 +67,7 @@ struct impl {
 
 	struct spa_log *log;
 	struct spa_loop *main_loop;
+	struct spa_system *main_system;
 
 	struct spa_hook_list hooks;
 
@@ -129,6 +133,14 @@ static void remove_device(struct impl *this, struct device *device)
 	*device = this->devices[--this->n_devices];
 }
 
+static void clear_devices(struct impl *this)
+{
+        uint32_t i;
+	for (i = 0; i < this->n_devices; i++)
+	        udev_device_unref(this->devices[i].dev);
+	this->n_devices = 0;
+}
+
 static uint32_t get_card_id(struct impl *this, struct udev_device *dev)
 {
 	const char *e, *str;
@@ -136,10 +148,10 @@ static uint32_t get_card_id(struct impl *this, struct udev_device *dev)
 	if (udev_device_get_property_value(dev, "ACP_IGNORE"))
 		return SPA_ID_INVALID;
 
-	if ((str = udev_device_get_property_value(dev, "SOUND_CLASS")) && strcmp(str, "modem") == 0)
+	if ((str = udev_device_get_property_value(dev, "SOUND_CLASS")) && spa_streq(str, "modem"))
 		return SPA_ID_INVALID;
 
-	if ((str = udev_device_get_property_value(dev, "SOUND_INITIALIZED")) == NULL)
+	if (udev_device_get_property_value(dev, "SOUND_INITIALIZED") == NULL)
 		return SPA_ID_INVALID;
 
 	if ((str = udev_device_get_property_value(dev, "DEVPATH")) == NULL)
@@ -169,7 +181,7 @@ static void unescape(const char *src, char *dst)
 {
 	const char *s;
 	char *d;
-	int h1, h2;
+	int h1 = 0, h2 = 0;
 	enum { TEXT, BACKSLASH, EX, FIRST } state = TEXT;
 
 	for (s = src, d = dst; *s; s++) {
@@ -234,44 +246,124 @@ static void unescape(const char *src, char *dst)
 	*d = 0;
 }
 
+static int check_device_available(struct impl *this, struct device *device, int *num_pcm)
+{
+	char path[PATH_MAX];
+	DIR *card = NULL, *pcm = NULL;
+	FILE *f;
+	char buf[16];
+	size_t sz;
+	struct dirent *entry, *entry_pcm;
+	int res;
+
+	/*
+	 * Check if some pcm devices of the card are busy.  Check it via /proc, as we
+	 * don't want to actually open any devices using alsa-lib (generates uncontrolled
+	 * number of inotify events), or replicate its subdevice logic.
+	 */
+
+	*num_pcm = 0;
+
+	spa_scnprintf(path, sizeof(path), "/proc/asound/card%u", (unsigned int)device->id);
+
+	if ((card = opendir(path)) == NULL)
+		goto done;
+
+	while ((errno = 0, entry = readdir(card)) != NULL) {
+		if (!(entry->d_type == DT_DIR &&
+				spa_strstartswith(entry->d_name, "pcm")))
+			continue;
+
+		/* Check device class */
+		spa_scnprintf(path, sizeof(path), "/sys/class/sound/pcmC%uD%s/pcm_class",
+				(unsigned int)device->id, entry->d_name+3);
+		f = fopen(path, "r");
+		if (f == NULL)
+			goto done;
+		sz = fread(buf, 1, sizeof(buf) - 1, f);
+		buf[sz] = '\0';
+		fclose(f);
+		if (spa_strstartswith(buf, "modem"))
+			continue;
+
+		/* Check busy status */
+		spa_scnprintf(path, sizeof(path), "/proc/asound/card%u/%s",
+				(unsigned int)device->id, entry->d_name);
+		if ((pcm = opendir(path)) == NULL)
+			goto done;
+
+		while ((errno = 0, entry_pcm = readdir(pcm)) != NULL) {
+			if (!(entry_pcm->d_type == DT_DIR &&
+					spa_strstartswith(entry_pcm->d_name, "sub")))
+				continue;
+
+			spa_scnprintf(path, sizeof(path), "/proc/asound/card%u/%s/%s/status",
+					(unsigned int)device->id, entry->d_name, entry_pcm->d_name);
+
+			f = fopen(path, "r");
+			if (f == NULL)
+				goto done;
+			sz = fread(buf, 1, 6, f);
+			buf[sz] = '\0';
+			fclose(f);
+
+			if (!spa_strstartswith(buf, "closed")) {
+				spa_log_debug(this->log, "card %u pcm device %s busy",
+						(unsigned int)device->id, entry->d_name);
+				errno = EBUSY;
+				goto done;
+			}
+			spa_log_debug(this->log, "card %u pcm device %s free",
+					(unsigned int)device->id, entry->d_name);
+		}
+		if (errno != 0)
+			goto done;
+
+		++*num_pcm;
+
+		closedir(pcm);
+		pcm = NULL;
+	}
+	if (errno != 0)
+		goto done;
+
+done:
+	res = -errno;
+	if (card)
+		closedir(card);
+	if (pcm)
+		closedir(pcm);
+	return res;
+}
+
 static int emit_object_info(struct impl *this, struct device *device)
 {
 	struct spa_device_object_info info;
 	uint32_t id = device->id;
 	struct udev_device *dev = device->dev;
-	snd_ctl_t *ctl_hndl;
 	const char *str;
 	char path[32], *cn = NULL, *cln = NULL;
 	struct spa_dict_item items[25];
 	uint32_t n_items = 0;
 	int res, pcm;
 
-	snprintf(path, sizeof(path), "hw:%u", id);
-	spa_log_debug(this->log, "open card %s", path);
+	/*
+	 * inotify close events under /dev/snd must not be emitted, except after setting
+	 * device->emitted to true. alsalib functions can be used after that.
+	 */
 
-	if ((res = snd_ctl_open(&ctl_hndl, path, 0)) < 0) {
-		spa_log_error(this->log, "can't open control for card %s: %s",
-				path, snd_strerror(res));
+	if ((res = check_device_available(this, device, &pcm)) < 0)
 		return res;
-	}
-
-	pcm = -1;
-	res = snd_ctl_pcm_next_device(ctl_hndl, &pcm);
-
-	spa_log_debug(this->log, "close card %s", path);
-	snd_ctl_close(ctl_hndl);
-
-	if (res < 0) {
-		spa_log_error(this->log, "error iterating devices: %s", snd_strerror(res));
-		device->ignored = true;
-		return res;
-	}
-	if (pcm < 0) {
+	if (pcm == 0) {
 		spa_log_debug(this->log, "no pcm devices for %s", path);
 		device->ignored = true;
 		return 0;
 	}
 
+	snprintf(path, sizeof(path), "hw:%u", id);
+	spa_log_debug(this->log, "emitting card %s", path);
+	device->emitted = true;
+
 	info = SPA_DEVICE_OBJECT_INFO_INIT();
 
 	info.type = SPA_TYPE_INTERFACE_Device;
@@ -323,7 +415,13 @@ static int emit_object_info(struct impl *this, struct device *device)
 		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_DEVICE_SUBSYSTEM, str);
 	}
 	if ((str = udev_device_get_property_value(dev, "ID_VENDOR_ID")) && *str) {
-		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_DEVICE_VENDOR_ID, str);
+		char *dec = alloca(6); /* 65535 is max */
+		int32_t val;
+
+		if (spa_atoi32(str, &val, 16)) {
+			snprintf(dec, 6, "%d", val);
+			items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_DEVICE_VENDOR_ID, dec);
+		}
 	}
 	str = udev_device_get_property_value(dev, "ID_VENDOR_FROM_DATABASE");
 	if (!(str && *str)) {
@@ -340,7 +438,13 @@ static int emit_object_info(struct impl *this, struct device *device)
 		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_DEVICE_VENDOR_NAME, str);
 	}
 	if ((str = udev_device_get_property_value(dev, "ID_MODEL_ID")) && *str) {
-		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_DEVICE_PRODUCT_ID, str);
+		char *dec = alloca(6); /* 65535 is max */
+		int32_t val;
+
+		if (spa_atoi32(str, &val, 16)) {
+			snprintf(dec, 6, "%d", val);
+			items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_DEVICE_PRODUCT_ID, dec);
+		}
 	}
 	str = udev_device_get_property_value(dev, "ID_MODEL_FROM_DATABASE");
 	if (!(str && *str)) {
@@ -365,7 +469,6 @@ static int emit_object_info(struct impl *this, struct device *device)
 	info.props = &SPA_DICT_INIT(items, n_items);
 
 	spa_device_emit_object_info(&this->hooks, id, &info);
-	device->emitted = true;
 	free(cn);
 	free(cln);
 
@@ -375,10 +478,13 @@ static int emit_object_info(struct impl *this, struct device *device)
 static bool check_access(struct impl *this, struct device *device)
 {
 	char path[128];
+	bool accessible;
 
 	snprintf(path, sizeof(path), "/dev/snd/controlC%u", device->id);
-	device->accessible = access(path, R_OK|W_OK) >= 0;
-	spa_log_debug(this->log, "%s accessible:%u", path, device->accessible);
+	accessible = access(path, R_OK|W_OK) >= 0;
+	if (accessible != device->accessible)
+		spa_log_debug(this->log, "%s accessible:%u", path, accessible);
+	device->accessible = accessible;
 
 	return device->accessible;
 }
@@ -388,6 +494,7 @@ static void process_device(struct impl *this, uint32_t action, struct udev_devic
 	uint32_t id;
 	struct device *device;
 	bool emitted;
+	int res;
 
 	if ((id = get_card_id(this, dev)) == SPA_ID_INVALID)
 		return;
@@ -404,7 +511,24 @@ static void process_device(struct impl *this, uint32_t action, struct udev_devic
 			return;
 		if (!check_access(this, device))
 			return;
-		emit_object_info(this, device);
+		res = emit_object_info(this, device);
+		if (res < 0) {
+			if (device->ignored)
+				spa_log_info(this->log, "ALSA card %u unavailable (%s): it is ignored",
+						device->id, spa_strerror(res));
+			else if (!device->unavailable)
+				spa_log_info(this->log, "ALSA card %u unavailable (%s): wait for it",
+						device->id, spa_strerror(res));
+			else
+				spa_log_debug(this->log, "ALSA card %u still unavailable (%s)",
+						device->id, spa_strerror(res));
+			device->unavailable = true;
+		} else {
+			if (device->unavailable)
+				spa_log_info(this->log, "ALSA card %u now available",
+						device->id);
+			device->unavailable = false;
+		}
 		break;
 
 	case ACTION_REMOVE:
@@ -415,6 +539,15 @@ static void process_device(struct impl *this, uint32_t action, struct udev_devic
 		if (emitted)
 			spa_device_emit_object_info(&this->hooks, id, NULL);
 		break;
+
+	case ACTION_DISABLE:
+		if (device == NULL)
+			return;
+		if (device->emitted) {
+			device->emitted = false;
+			spa_device_emit_object_info(&this->hooks, id, NULL);
+		}
+		break;
 	}
 }
 
@@ -449,22 +582,33 @@ static void impl_on_notify_events(struct spa_source *source)
 		if (len <= 0)
 			break;
 
-		e = SPA_MEMBER(&buf, len, void);
+		e = SPA_PTROFF(&buf, len, void);
 
 		for (p = &buf; p < e;
-		    p = SPA_MEMBER(p, sizeof(struct inotify_event) + event->len, void)) {
+		    p = SPA_PTROFF(p, sizeof(struct inotify_event) + event->len, void)) {
 			unsigned int id;
 			struct device *device;
 
 			event = (const struct inotify_event *) p;
 
-			if ((event->mask & IN_ATTRIB)) {
-				if (sscanf(event->name, "controlC%u", &id) != 1)
+			/* Device becomes accessible or not busy */
+			if ((event->mask & (IN_ATTRIB | IN_CLOSE_WRITE))) {
+				bool access;
+
+				if ((event->mask & IN_ATTRIB) &&
+						spa_strstartswith(event->name, "pcm"))
+					continue;
+				if (sscanf(event->name, "controlC%u", &id) != 1 &&
+						sscanf(event->name, "pcmC%uD", &id) != 1)
 					continue;
 				if ((device = find_device(this, id)) == NULL)
 					continue;
-				if (!device->emitted)
+
+				access = check_access(this, device);
+				if (access && !device->emitted)
 					process_device(this, ACTION_ADD, device->dev);
+				else if (!access && device->emitted)
+					process_device(this, ACTION_DISABLE, device->dev);
 			}
 			/* /dev/snd/ might have been removed */
 			if ((event->mask & (IN_DELETE_SELF | IN_MOVE_SELF)))
@@ -526,9 +670,9 @@ static void impl_on_fd_events(struct spa_source *source)
 
 	start_inotify(this);
 
-	if (strcmp(action, "change") == 0) {
+	if (spa_streq(action, "change")) {
 		process_device(this, ACTION_ADD, dev);
-	} else if (strcmp(action, "remove") == 0) {
+	} else if (spa_streq(action, "remove")) {
 		process_device(this, ACTION_REMOVE, dev);
 	}
 	udev_device_unref(dev);
@@ -551,7 +695,7 @@ static int start_monitor(struct impl *this)
 
 	this->source.func = impl_on_fd_events;
 	this->source.data = this;
-	this->source.fd = udev_monitor_get_fd(this->umonitor);;
+	this->source.fd = udev_monitor_get_fd(this->umonitor);
 	this->source.mask = SPA_IO_IN | SPA_IO_ERR;
 
 	spa_log_debug(this->log, "monitor %p", this->umonitor);
@@ -568,6 +712,8 @@ static int stop_monitor(struct impl *this)
 	if (this->umonitor == NULL)
 		return 0;
 
+        clear_devices (this);
+
 	spa_loop_remove_source(this->main_loop, &this->source);
 	udev_monitor_unref(this->umonitor);
 	this->umonitor = NULL;
@@ -614,12 +760,13 @@ static const struct spa_dict_item device_info_items[] = {
 
 static void emit_device_info(struct impl *this, bool full)
 {
+	uint64_t old = full ? this->info.change_mask : 0;
 	if (full)
 		this->info.change_mask = this->info_all;
 	if (this->info.change_mask) {
 		this->info.props = &SPA_DICT_INIT_ARRAY(device_info_items);
 		spa_device_emit_info(&this->hooks, &this->info);
-		this->info.change_mask = 0;
+		this->info.change_mask = old;
 	}
 }
 
@@ -650,10 +797,10 @@ impl_device_add_listener(void *object, struct spa_hook *listener,
 
 	emit_device_info(this, true);
 
-	if ((res = enum_devices(this)) < 0)
+	if ((res = start_monitor(this)) < 0)
 		return res;
 
-	if ((res = start_monitor(this)) < 0)
+	if ((res = enum_devices(this)) < 0)
 		return res;
 
         spa_hook_list_join(&this->hooks, &save);
@@ -678,7 +825,7 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 
 	this = (struct impl *) handle;
 
-	if (strcmp(type, SPA_TYPE_INTERFACE_Device) == 0)
+	if (spa_streq(type, SPA_TYPE_INTERFACE_Device))
 		*interface = &this->device;
 	else
 		return -ENOENT;
@@ -721,12 +868,18 @@ impl_init(const struct spa_handle_factory *factory,
 	this->notify.fd = -1;
 
 	this->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
+	alsa_log_topic_init(this->log);
 	this->main_loop = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Loop);
+	this->main_system = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_System);
 
 	if (this->main_loop == NULL) {
 		spa_log_error(this->log, "a main-loop is needed");
 		return -EINVAL;
 	}
+	if (this->main_system == NULL) {
+		spa_log_error(this->log, "a main-system is needed");
+		return -EINVAL;
+	}
 	spa_hook_list_init(&this->hooks);
 
 	this->device.iface = SPA_INTERFACE_INIT(
@@ -741,7 +894,7 @@ impl_init(const struct spa_handle_factory *factory,
 
 	if (info) {
 		if ((str = spa_dict_lookup(info, "alsa.use-acp")) != NULL)
-			this->use_acp = strcmp(str, "true") == 0 || atoi(str) != 0;
+			this->use_acp = spa_atob(str);
 	}
 
 	return 0;
