@@ -56,8 +56,8 @@ static struct spa_log_topic log_topic = SPA_LOG_TOPIC(0, "spa.bluez5.native");
 
 #define PROP_KEY_HEADSET_ROLES "bluez5.headset-roles"
 
-#define HFP_CODEC_SWITCH_INITIAL_TIMEOUT_MSEC 2000
-#define HFP_CODEC_SWITCH_TIMEOUT_MSEC 5000
+#define HFP_CODEC_SWITCH_INITIAL_TIMEOUT_MSEC 5000
+#define HFP_CODEC_SWITCH_TIMEOUT_MSEC 20000
 
 enum {
 	HFP_AG_INITIAL_CODEC_SETUP_NONE = 0,
@@ -137,6 +137,8 @@ struct rfcomm {
 	unsigned int msbc_supported_by_hfp:1;
 	unsigned int hfp_ag_switching_codec:1;
 	unsigned int hfp_ag_initial_codec_setup:2;
+	unsigned int cind_call_active:1;
+	unsigned int cind_call_notify:1;
 	enum hfp_hf_state hf_state;
 	enum hsp_hs_state hs_state;
 	unsigned int codec;
@@ -199,12 +201,6 @@ static struct spa_bt_transport *_transport_create(struct rfcomm *rfcomm)
 	td = t->user_data;
 	td->rfcomm = rfcomm;
 
-	for (int i = 0; i < SPA_BT_VOLUME_ID_TERM ; ++i) {
-		rfcomm->volumes[i].hw_volume = SPA_BT_VOLUME_INVALID;
-		t->volumes[i].active = rfcomm->volumes[i].active;
-		t->volumes[i].hw_volume_max = SPA_BT_VOLUME_HS_MAX;
-	}
-
 	if (t->profile & SPA_BT_PROFILE_HEADSET_AUDIO_GATEWAY) {
 		t->volumes[SPA_BT_VOLUME_ID_RX].volume = DEFAULT_AG_VOLUME;
 		t->volumes[SPA_BT_VOLUME_ID_TX].volume = DEFAULT_AG_VOLUME;
@@ -213,6 +209,14 @@ static struct spa_bt_transport *_transport_create(struct rfcomm *rfcomm)
 		t->volumes[SPA_BT_VOLUME_ID_TX].volume = DEFAULT_TX_VOLUME;
 	}
 
+	for (int i = 0; i < SPA_BT_VOLUME_ID_TERM ; ++i) {
+		t->volumes[i].active = rfcomm->volumes[i].active;
+		t->volumes[i].hw_volume_max = SPA_BT_VOLUME_HS_MAX;
+		if (rfcomm->volumes[i].active && rfcomm->volumes[i].hw_volume != SPA_BT_VOLUME_INVALID)
+			t->volumes[i].volume =
+				spa_bt_volume_hw_to_linear(rfcomm->volumes[i].hw_volume, t->volumes[i].hw_volume_max);
+	}
+
 	spa_bt_transport_add_listener(t, &rfcomm->transport_listener, &transport_events, rfcomm);
 
 finish:
@@ -353,9 +357,8 @@ static void rfcomm_emit_volume_changed(struct rfcomm *rfcomm, int id, int hw_vol
 }
 
 #ifdef HAVE_BLUEZ_5_BACKEND_HSP_NATIVE
-static bool rfcomm_hsp_ag(struct spa_source *source, char* buf)
+static bool rfcomm_hsp_ag(struct rfcomm *rfcomm, char* buf)
 {
-	struct rfcomm *rfcomm = source->data;
 	struct impl *backend = rfcomm->backend;
 	unsigned int gain, dummy;
 
@@ -389,9 +392,8 @@ static bool rfcomm_hsp_ag(struct spa_source *source, char* buf)
 	return true;
 }
 
-static bool rfcomm_send_volume_cmd(struct spa_source *source, int id)
+static bool rfcomm_send_volume_cmd(struct rfcomm *rfcomm, int id)
 {
-	struct rfcomm *rfcomm = source->data;
 	struct spa_bt_transport_volume *t_volume;
 	const char *format;
 	int hw_volume;
@@ -419,9 +421,8 @@ static bool rfcomm_send_volume_cmd(struct spa_source *source, int id)
 	return true;
 }
 
-static bool rfcomm_hsp_hs(struct spa_source *source, char* buf)
+static bool rfcomm_hsp_hs(struct rfcomm *rfcomm, char* buf)
 {
-	struct rfcomm *rfcomm = source->data;
 	struct impl *backend = rfcomm->backend;
 	unsigned int gain;
 
@@ -444,14 +445,14 @@ static bool rfcomm_hsp_hs(struct spa_source *source, char* buf)
 		} else {
 			spa_log_debug(backend->log, "RFCOMM receive unsupported VGM gain: %s", buf);
 		}
-	} if (strncmp(buf, "\r\nOK\r\n", 6) == 0) {
+	} else if (spa_strstartswith(buf, "\r\nOK\r\n")) {
 		if (rfcomm->hs_state == hsp_hs_init2) {
-			if (rfcomm_send_volume_cmd(&rfcomm->source, SPA_BT_VOLUME_ID_RX))
+			if (rfcomm_send_volume_cmd(rfcomm, SPA_BT_VOLUME_ID_RX))
 				rfcomm->hs_state = hsp_hs_vgs;
 			else
 				rfcomm->hs_state = hsp_hs_init1;
 		} else if (rfcomm->hs_state == hsp_hs_vgs) {
-			if (rfcomm_send_volume_cmd(&rfcomm->source, SPA_BT_VOLUME_ID_TX))
+			if (rfcomm_send_volume_cmd(rfcomm, SPA_BT_VOLUME_ID_TX))
 				rfcomm->hs_state = hsp_hs_vgm;
 			else
 				rfcomm->hs_state = hsp_hs_init1;
@@ -551,11 +552,10 @@ fail:
 
 #ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
 
-static int sco_create_socket(struct impl *backend, struct spa_bt_adapter *adapter, bool msbc);
-
 static bool device_supports_required_mSBC_transport_modes(
-		struct impl *backend, struct spa_bt_device *device) {
-	int sock;
+		struct impl *backend, struct spa_bt_device *device)
+{
+	int res;
 	bool msbc_ok, msbc_alt1_ok;
 	uint32_t bt_features;
 
@@ -576,43 +576,20 @@ static bool device_supports_required_mSBC_transport_modes(
 	if (!msbc_ok && !msbc_alt1_ok)
 		return false;
 
-	/*
-	 * Check if adapter supports BT_VOICE_TRANSPARENT. Do this without
-	 * directly probing HCI properties.
-	 */
-	sock = sco_create_socket(backend, device->adapter, true);
-	if (sock < 0) {
+	res = spa_bt_adapter_has_msbc(device->adapter);
+	if (res < 0) {
+		spa_log_warn(backend->log,
+				"adapter %s: failed to determine msbc/esco capability (%d)",
+				device->adapter->path, res);
+	} else if (res == 0) {
+		spa_log_info(backend->log,
+				"adapter %s: no msbc/esco transport",
+				device->adapter->path);
 		return false;
 	} else {
-		struct sockaddr_sco addr;
-		socklen_t len;
-		int res;
-
-		/* Connect to non-existent address */
-		len = sizeof(addr);
-		memset(&addr, 0, len);
-		addr.sco_family = AF_BLUETOOTH;
-		bacpy(&addr.sco_bdaddr, BDADDR_LOCAL);
-
-		spa_log_debug(backend->log, "connect to determine adapter msbc support...");
-
-		/* Linux kernel code checks for features needed for BT_VOICE_TRANSPARENT
-		 * among the first checks it does, and fails with EOPNOTSUPP if not
-		 * supported. The connection generally timeouts, so set it
-		 * nonblocking since we are just checking.
-		 */
-		fcntl(sock, F_SETFL, O_NONBLOCK);
-		res = connect(sock, (struct sockaddr *) &addr, len);
-		if (res < 0)
-			res = errno;
-		else
-			res = 0;
-		close(sock);
-
-		spa_log_debug(backend->log, "determined adapter-msbc:%d res:%d",
-				(res != EOPNOTSUPP), res);
-		if (res == EOPNOTSUPP)
-			return false;
+		spa_log_debug(backend->log,
+				"adapter %s: has msbc/esco transport",
+				device->adapter->path);
 	}
 
 	/* Check if USB ALT6 is really available on the device */
@@ -640,19 +617,72 @@ static bool device_supports_required_mSBC_transport_modes(
 
 static int codec_switch_start_timer(struct rfcomm *rfcomm, int timeout_msec);
 
-static bool rfcomm_hfp_ag(struct spa_source *source, char* buf)
+static void process_iphoneaccev_indicator(struct rfcomm *rfcomm, unsigned int key, unsigned int value)
+{
+	struct impl *backend = rfcomm->backend;
+
+	spa_log_debug(backend->log, "key:%u value:%u", key, value);
+
+	if (key == SPA_BT_HFP_HF_IPHONEACCEV_KEY_BATTERY_LEVEL) {
+		// Battery level is reported in range of 0-9, convert to 10-100%
+		uint8_t level = (SPA_CLAMP(value, 0u, 9u) + 1) * 10;
+		spa_log_debug(backend->log, "battery level: %d%%", (int) level);
+
+		// TODO: report without Battery Provider (using props)
+		spa_bt_device_report_battery_level(rfcomm->device, level);
+	} else {
+		spa_log_warn(backend->log, "unknown AT+IPHONEACCEV key:%u value:%u", key, value);
+	}
+}
+
+static void process_hfp_hf_indicator(struct rfcomm *rfcomm, unsigned int indicator, unsigned int value)
+{
+	struct impl *backend = rfcomm->backend;
+
+	spa_log_debug(backend->log, "indicator:%u value:%u", indicator, value);
+
+	if (indicator == SPA_BT_HFP_HF_INDICATOR_BATTERY_LEVEL) {
+		// Battery level is reported in range 0-100
+		spa_log_debug(backend->log, "battery level: %u%%", value);
+
+		if (value <= 100) {
+			// TODO: report without Battery Provider (using props)
+			spa_bt_device_report_battery_level(rfcomm->device, value);
+		} else {
+			spa_log_warn(backend->log, "battery HF indicator %u outside of range [0, 100]: %u", indicator, value);
+		}
+	} else {
+		spa_log_warn(backend->log, "unknown HF indicator:%u value:%u", indicator, value);
+	}
+}
+
+static void rfcomm_hfp_ag_set_cind(struct rfcomm *rfcomm, bool call_active)
+{
+	if (rfcomm->profile != SPA_BT_PROFILE_HFP_HF)
+		return;
+
+	if (call_active == rfcomm->cind_call_active)
+		return;
+
+	rfcomm->cind_call_active = call_active;
+
+	if (!rfcomm->cind_call_notify)
+		return;
+
+	rfcomm_send_reply(rfcomm, "+CIEV: 2,%d", rfcomm->cind_call_active);
+}
+
+static bool rfcomm_hfp_ag(struct rfcomm *rfcomm, char* buf)
 {
-	struct rfcomm *rfcomm = source->data;
 	struct impl *backend = rfcomm->backend;
 	unsigned int features;
 	unsigned int gain;
-	unsigned int len;
+	unsigned int count, r;
 	unsigned int selected_codec;
 	unsigned int indicator;
 	unsigned int indicator_value;
 	int xapl_vendor;
 	int xapl_product;
-	int xapl_version;
 	int xapl_features;
 
 	if (sscanf(buf, "AT+BRSF=%u", &features) == 1) {
@@ -695,7 +725,7 @@ static bool rfcomm_hfp_ag(struct spa_source *source, char* buf)
 		ag_features |= SPA_BT_HFP_AG_FEATURE_HF_INDICATORS;
 		rfcomm_send_reply(rfcomm, "+BRSF: %u", ag_features);
 		rfcomm_send_reply(rfcomm, "OK");
-	} else if (strncmp(buf, "AT+BAC=", 7) == 0) {
+	} else if (spa_strstartswith(buf, "AT+BAC=")) {
 		/* retrieve supported codecs */
 		/* response has the form AT+BAC=<codecID1>,<codecID2>,<codecIDx>
 		   strategy: split the string into tokens */
@@ -723,16 +753,24 @@ static bool rfcomm_hfp_ag(struct spa_source *source, char* buf)
 		}
 
 		rfcomm_send_reply(rfcomm, "OK");
-	} else if (strncmp(buf, "AT+CIND=?", 9) == 0) {
+	} else if (spa_strstartswith(buf, "AT+CIND=?")) {
 		rfcomm_send_reply(rfcomm, "+CIND:(\"service\",(0-1)),(\"call\",(0-1)),(\"callsetup\",(0-3)),(\"callheld\",(0-2))");
 		rfcomm_send_reply(rfcomm, "OK");
-	} else if (strncmp(buf, "AT+CIND?", 8) == 0) {
-		rfcomm_send_reply(rfcomm, "+CIND: 0,0,0,0");
+	} else if (spa_strstartswith(buf, "AT+CIND?")) {
+		rfcomm_send_reply(rfcomm, "+CIND: 0,%d,0,0", rfcomm->cind_call_active);
 		rfcomm_send_reply(rfcomm, "OK");
-	} else if (strncmp(buf, "AT+CMER", 7) == 0) {
+	} else if (spa_strstartswith(buf, "AT+CMER")) {
+		int mode, keyp, disp, ind;
+
 		rfcomm->slc_configured = true;
 		rfcomm_send_reply(rfcomm, "OK");
 
+		rfcomm->cind_call_active = false;
+		if (sscanf(buf, "AT+CMER= %d , %d , %d , %d", &mode, &keyp, &disp, &ind) == 4)
+			rfcomm->cind_call_notify = ind ? true : false;
+		else
+			rfcomm->cind_call_notify = false;
+
 		/* switch codec to mSBC by sending unsolicited +BCS message */
 		if (rfcomm->codec_negotiation_supported && rfcomm->msbc_supported_by_hfp) {
 			spa_log_debug(backend->log, "RFCOMM initial codec setup");
@@ -750,7 +788,6 @@ static bool rfcomm_hfp_ag(struct spa_source *source, char* buf)
 				rfcomm_emit_volume_changed(rfcomm, -1, SPA_BT_VOLUME_INVALID);
 			}
 		}
-
 	} else if (!rfcomm->slc_configured) {
 		spa_log_warn(backend->log, "RFCOMM receive command before SLC completed: %s", buf);
 		rfcomm_send_reply(rfcomm, "ERROR");
@@ -794,6 +831,11 @@ static bool rfcomm_hfp_ag(struct spa_source *source, char* buf)
 		rfcomm_send_reply(rfcomm, "OK");
 		if (was_switching_codec)
 			spa_bt_device_emit_codec_switched(rfcomm->device, 0);
+	} else if (spa_strstartswith(buf, "AT+BIA=")) {
+		/* We only support 'call' indicator, which HFP 4.35.1 defines as
+		   always active (assuming CMER enabled it), so we don't need to
+		   parse anything here. */
+		rfcomm_send_reply(rfcomm, "OK");
 	} else if (sscanf(buf, "AT+VGM=%u", &gain) == 1) {
 		if (gain <= SPA_BT_VOLUME_HS_MAX) {
 			if (!rfcomm->broken_mic_hw_volume)
@@ -823,57 +865,29 @@ static bool rfcomm_hfp_ag(struct spa_source *source, char* buf)
 		// is supported
 		rfcomm_send_reply(rfcomm, "OK");
 	} else if (sscanf(buf, "AT+BIEV=%u,%u", &indicator, &indicator_value) == 2) {
-		if (indicator == SPA_BT_HFP_HF_INDICATOR_BATTERY_LEVEL) {
-			// Battery level is reported in range 0-100
-			spa_log_debug(backend->log, "battery level: %u%%", indicator_value);
-
-			if (indicator_value <= 100) {
-				// TODO: report without Battery Provider (using props)
-				spa_bt_device_report_battery_level(rfcomm->device, indicator_value);
-			} else {
-				spa_log_warn(backend->log, "battery HF indicator %u outside of range [0, 100]: %u", indicator, indicator_value);
-			}
-		} else {
-			spa_log_warn(backend->log, "unknown HF indicator: %u", indicator);
-		}
-	} else if (sscanf(buf, "AT+XAPL=%04x-%04x-%04x,%u", &xapl_vendor, &xapl_product, &xapl_version, &xapl_features) == 4) {
+		process_hfp_hf_indicator(rfcomm, indicator, indicator_value);
+	} else if (sscanf(buf, "AT+XAPL=%04x-%04x-%*[^,],%u", &xapl_vendor, &xapl_product, &xapl_features) == 3) {
 		if (xapl_features & SPA_BT_HFP_HF_XAPL_FEATURE_BATTERY_REPORTING) {
 			/* claim, that we support battery status reports */
 			rfcomm_send_reply(rfcomm, "+XAPL=iPhone,%u", SPA_BT_HFP_HF_XAPL_FEATURE_BATTERY_REPORTING);
 		}
 		rfcomm_send_reply(rfcomm, "OK");
-	} else if (sscanf(buf, "AT+IPHONEACCEV=%u", &len) == 1) {
-		unsigned int i;
-		int k, v;
-
-		if (len < 1 || len > 100)
+	} else if (sscanf(buf, "AT+IPHONEACCEV=%u%n", &count, &r) == 1) {
+		if (count < 1 || count > 100)
 			return false;
 
-		for (i = 1; i <= len; i++) {
-			buf = strchr(buf, ',');
-			if (buf == NULL)
-				return false;
-			++buf;
-			if (sscanf(buf, "%d", &k) != 1)
+		buf += r;
+
+		for (unsigned int i = 0; i < count; i++) {
+			unsigned int key, value;
+
+			if (sscanf(buf, " , %u , %u%n", &key, &value, &r) != 2)
 				return false;
 
-			buf = strchr(buf, ',');
-			if (buf == NULL)
-				return false;
-			++buf;
-			if (sscanf(buf, "%d", &v) != 1)
-				return false;
-
-			if (k == SPA_BT_HFP_HF_IPHONEACCEV_KEY_BATTERY_LEVEL) {
-				// Battery level is reported in range of 0-9, convert to 0-100%
-				uint8_t level = (SPA_CLAMP(v, 0, 9) + 1) * 10;
-				spa_log_debug(backend->log, "battery level: %d%%", (int)level);
-
-				// TODO: report without Battery Provider (using props)
-				spa_bt_device_report_battery_level(rfcomm->device, level);
-			}
+			process_iphoneaccev_indicator(rfcomm, key, value);
+			buf += r;
 		}
-	} else if (strncmp(buf, "AT+APLSIRI?", 11) == 0) {
+	} else if (spa_strstartswith(buf, "AT+APLSIRI?")) {
 		// This command is sent when we activate Apple extensions
 		rfcomm_send_reply(rfcomm, "OK");
 	} else {
@@ -883,11 +897,10 @@ static bool rfcomm_hfp_ag(struct spa_source *source, char* buf)
 	return true;
 }
 
-static bool rfcomm_hfp_hf(struct spa_source *source, char* buf)
+static bool rfcomm_hfp_hf(struct rfcomm *rfcomm, char* buf)
 {
 	static const char separators[] = "\r\n:";
 
-	struct rfcomm *rfcomm = source->data;
 	struct impl *backend = rfcomm->backend;
 	unsigned int features;
 	unsigned int gain;
@@ -897,14 +910,14 @@ static bool rfcomm_hfp_hf(struct spa_source *source, char* buf)
 	token = strtok(buf, separators);
 	while (token != NULL)
 	{
-		if (strncmp(token, "+BRSF", 5) == 0) {
+		if (spa_strstartswith(token, "+BRSF")) {
 			/* get next token */
 			token = strtok(NULL, separators);
 			features = atoi(token);
 			if (((features & (SPA_BT_HFP_AG_FEATURE_CODEC_NEGOTIATION)) != 0) &&
 			    rfcomm->msbc_supported_by_hfp)
 				rfcomm->codec_negotiation_supported = true;
-		} else if (strncmp(token, "+BCS", 4) == 0 && rfcomm->codec_negotiation_supported) {
+		} else if (spa_strstartswith(token, "+BCS") && rfcomm->codec_negotiation_supported) {
 			/* get next token */
 			token = strtok(NULL, separators);
 			selected_codec = atoi(token);
@@ -933,10 +946,10 @@ static bool rfcomm_hfp_hf(struct spa_source *source, char* buf)
 					}
 				}
 			}
-		} else if (strncmp(token, "+CIND", 5) == 0) {
+		} else if (spa_strstartswith(token, "+CIND")) {
 			/* get next token and discard it */
 			token = strtok(NULL, separators);
-		} else if (strncmp(token, "+VGM", 4) == 0) {
+		} else if (spa_strstartswith(token, "+VGM")) {
 			/* get next token */
 			token = strtok(NULL, separators);
 			gain = atoi(token);
@@ -946,7 +959,7 @@ static bool rfcomm_hfp_hf(struct spa_source *source, char* buf)
 			} else {
 				spa_log_debug(backend->log, "RFCOMM receive unsupported VGM gain: %s", token);
 			}
-		} else if (strncmp(token, "+VGS", 4) == 0) {
+		} else if (spa_strstartswith(token, "+VGS")) {
 			/* get next token */
 			token = strtok(NULL, separators);
 			gain = atoi(token);
@@ -956,7 +969,7 @@ static bool rfcomm_hfp_hf(struct spa_source *source, char* buf)
 			} else {
 				spa_log_debug(backend->log, "RFCOMM receive unsupported VGS gain: %s", token);
 			}
-		} else if (strncmp(token, "OK", 5) == 0) {
+		} else if (spa_strstartswith(token, "OK")) {
 			switch(rfcomm->hf_state) {
 				case hfp_hf_brsf:
 					if (rfcomm->codec_negotiation_supported) {
@@ -993,16 +1006,16 @@ static bool rfcomm_hfp_hf(struct spa_source *source, char* buf)
 						}
 					}
 					/* Report volume on SLC establishment */
-					if (rfcomm_send_volume_cmd(source, SPA_BT_VOLUME_ID_RX))
+					if (rfcomm_send_volume_cmd(rfcomm, SPA_BT_VOLUME_ID_RX))
 						rfcomm->hf_state = hfp_hf_vgs;
 					break;
 				case hfp_hf_slc2:
-					if (rfcomm_send_volume_cmd(source, SPA_BT_VOLUME_ID_RX))
+					if (rfcomm_send_volume_cmd(rfcomm, SPA_BT_VOLUME_ID_RX))
 						rfcomm->hf_state = hfp_hf_vgs;
 					break;
 				case hfp_hf_vgs:
 					rfcomm->hf_state = hfp_hf_slc1;
-					if (rfcomm_send_volume_cmd(source, SPA_BT_VOLUME_ID_TX))
+					if (rfcomm_send_volume_cmd(rfcomm, SPA_BT_VOLUME_ID_TX))
 						rfcomm->hf_state = hfp_hf_vgm;
 					break;
 				default:
@@ -1042,25 +1055,32 @@ static void rfcomm_event(struct spa_source *source)
 		buf[len] = 0;
 		spa_log_debug(backend->log, "RFCOMM << %s", buf);
 
+		switch (rfcomm->profile) {
 #ifdef HAVE_BLUEZ_5_BACKEND_HSP_NATIVE
-		if (rfcomm->profile == SPA_BT_PROFILE_HSP_HS)
-			res = rfcomm_hsp_ag(source, buf);
-		else if (rfcomm->profile == SPA_BT_PROFILE_HSP_AG)
-			res = rfcomm_hsp_hs(source, buf);
+		case SPA_BT_PROFILE_HSP_HS:
+			res = rfcomm_hsp_ag(rfcomm, buf);
+			break;
+		case SPA_BT_PROFILE_HSP_AG:
+			res = rfcomm_hsp_hs(rfcomm, buf);
+			break;
 #endif
 #ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
-		if (rfcomm->profile == SPA_BT_PROFILE_HFP_HF)
-			res = rfcomm_hfp_ag(source, buf);
-		else if (rfcomm->profile == SPA_BT_PROFILE_HFP_AG)
-			res = rfcomm_hfp_hf(source, buf);
+		case SPA_BT_PROFILE_HFP_HF:
+			res = rfcomm_hfp_ag(rfcomm, buf);
+			break;
+		case SPA_BT_PROFILE_HFP_AG:
+			res = rfcomm_hfp_hf(rfcomm, buf);
+			break;
 #endif
+		default:
+			break;
+		}
 
 		if (!res) {
-			spa_log_debug(backend->log, "RFCOMM receive unsupported command: %s", buf);
+			spa_log_debug(backend->log, "RFCOMM received unsupported command: %s", buf);
 			rfcomm_send_reply(rfcomm, "ERROR");
 		}
 	}
-	return;
 }
 
 static int sco_create_socket(struct impl *backend, struct spa_bt_adapter *adapter, bool msbc)
@@ -1112,6 +1132,7 @@ static int sco_do_connect(struct spa_bt_transport *t)
 {
 	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
 	struct spa_bt_device *d = t->device;
+	struct transport_data *td = t->user_data;
 	struct sockaddr_sco addr;
 	socklen_t len;
 	int err;
@@ -1146,6 +1167,21 @@ again:
 		goto again;
 	} else if (err < 0 && !(errno == EAGAIN || errno == EINPROGRESS)) {
 		spa_log_error(backend->log, "connect(): %s", strerror(errno));
+#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
+		if (errno == EOPNOTSUPP && t->codec == HFP_AUDIO_CODEC_MSBC &&
+				td->rfcomm->msbc_supported_by_hfp) {
+			/* Adapter doesn't support msbc. Renegotiate. */
+			d->adapter->msbc_probed = true;
+			d->adapter->has_msbc = false;
+			td->rfcomm->msbc_supported_by_hfp = false;
+			if (t->profile == SPA_BT_PROFILE_HFP_HF) {
+				td->rfcomm->hfp_ag_switching_codec = true;
+				rfcomm_send_reply(td->rfcomm, "+BCS: 1");
+			} else if (t->profile == SPA_BT_PROFILE_HFP_AG) {
+				rfcomm_send_cmd(td->rfcomm, "AT+BAC=1");
+			}
+		}
+#endif
 		goto fail_close;
 	}
 
@@ -1159,6 +1195,7 @@ fail_close:
 static int sco_acquire_cb(void *data, bool optional)
 {
 	struct spa_bt_transport *t = data;
+	struct transport_data *td = t->user_data;
 	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
 	int sock;
 	socklen_t len;
@@ -1173,6 +1210,10 @@ static int sco_acquire_cb(void *data, bool optional)
 	if (sock < 0)
 		goto fail;
 
+#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
+	rfcomm_hfp_ag_set_cind(td->rfcomm, true);
+#endif
+
 	t->fd = sock;
 
 	/* Fallback value */
@@ -1204,10 +1245,15 @@ fail:
 static int sco_release_cb(void *data)
 {
 	struct spa_bt_transport *t = data;
+	struct transport_data *td = t->user_data;
 	struct impl *backend = SPA_CONTAINER_OF(t->backend, struct impl, this);
 
 	spa_log_info(backend->log, "Transport %s released", t->path);
 
+#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
+	rfcomm_hfp_ag_set_cind(td->rfcomm, false);
+#endif
+
 	if (t->sco_io) {
 		spa_bt_sco_io_destroy(t->sco_io);
 		t->sco_io = NULL;
@@ -1245,10 +1291,10 @@ static void sco_listen_event(struct spa_source *source)
 {
 	struct impl *backend = source->data;
 	struct sockaddr_sco addr;
-	socklen_t optlen;
+	socklen_t addrlen;
 	int sock = -1;
 	char local_address[18], remote_address[18];
-	struct rfcomm *rfcomm, *rfcomm_tmp;
+	struct rfcomm *rfcomm;
 	struct spa_bt_transport *t = NULL;
 	struct transport_data *td;
 
@@ -1258,10 +1304,10 @@ static void sco_listen_event(struct spa_source *source)
 	}
 
 	memset(&addr, 0, sizeof(addr));
-	optlen = sizeof(addr);
+	addrlen = sizeof(addr);
 
 	spa_log_debug(backend->log, "doing accept");
-	sock = accept(source->fd, (struct sockaddr *) &addr, &optlen);
+	sock = accept(source->fd, (struct sockaddr *) &addr, &addrlen);
 	if (sock < 0) {
 		if (errno != EAGAIN)
 			spa_log_error(backend->log, "SCO accept(): %s", strerror(errno));
@@ -1271,9 +1317,9 @@ static void sco_listen_event(struct spa_source *source)
 	ba2str(&addr.sco_bdaddr, remote_address);
 
 	memset(&addr, 0, sizeof(addr));
-	optlen = sizeof(addr);
+	addrlen = sizeof(addr);
 
-	if (getsockname(sock, (struct sockaddr *) &addr, &optlen) < 0) {
+	if (getsockname(sock, (struct sockaddr *) &addr, &addrlen) < 0) {
 		spa_log_error(backend->log, "SCO getsockname(): %s", strerror(errno));
 		goto fail;
 	}
@@ -1281,7 +1327,7 @@ static void sco_listen_event(struct spa_source *source)
 	ba2str(&addr.sco_bdaddr, local_address);
 
 	/* Find transport for local and remote address */
-	spa_list_for_each_safe(rfcomm, rfcomm_tmp, &backend->rfcomm_list, link) {
+	spa_list_for_each(rfcomm, &backend->rfcomm_list, link) {
 		if (rfcomm->transport && spa_streq(rfcomm->transport->device->address, remote_address) &&
 		    spa_streq(rfcomm->transport->device->adapter->address, local_address)) {
 					t = rfcomm->transport;
@@ -1345,12 +1391,12 @@ static void sco_listen_event(struct spa_source *source)
 
 	/* Report initial volume to remote */
 	if (t->profile == SPA_BT_PROFILE_HSP_AG) {
-		if (rfcomm_send_volume_cmd(&rfcomm->source, SPA_BT_VOLUME_ID_RX))
+		if (rfcomm_send_volume_cmd(rfcomm, SPA_BT_VOLUME_ID_RX))
 			rfcomm->hs_state = hsp_hs_vgs;
 		else
 			rfcomm->hs_state = hsp_hs_init1;
 	} else if (t->profile == SPA_BT_PROFILE_HFP_AG) {
-		if (rfcomm_send_volume_cmd(&rfcomm->source, SPA_BT_VOLUME_ID_RX))
+		if (rfcomm_send_volume_cmd(rfcomm, SPA_BT_VOLUME_ID_RX))
 			rfcomm->hf_state = hfp_hf_vgs;
 		else
 			rfcomm->hf_state = hfp_hf_slc1;
@@ -1627,13 +1673,34 @@ static const struct spa_bt_device_events device_events = {
 	.destroy = device_destroy,
 };
 
+static enum spa_bt_profile path_to_profile(const char *path)
+{
+#ifdef HAVE_BLUEZ_5_BACKEND_HSP_NATIVE
+	if (spa_streq(path, PROFILE_HSP_AG))
+		return SPA_BT_PROFILE_HSP_HS;
+
+	if (spa_streq(path, PROFILE_HSP_HS))
+		return SPA_BT_PROFILE_HSP_AG;
+#endif
+
+#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
+	if (spa_streq(path, PROFILE_HFP_AG))
+		return SPA_BT_PROFILE_HFP_HF;
+
+	if (spa_streq(path, PROFILE_HFP_HF))
+		return SPA_BT_PROFILE_HFP_AG;
+#endif
+
+	return SPA_BT_PROFILE_NULL;
+}
+
 static DBusHandlerResult profile_new_connection(DBusConnection *conn, DBusMessage *m, void *userdata)
 {
 	struct impl *backend = userdata;
 	DBusMessage *r;
 	DBusMessageIter it[5];
 	const char *handler, *path;
-	enum spa_bt_profile profile = SPA_BT_PROFILE_NULL;
+	enum spa_bt_profile profile;
 	struct rfcomm *rfcomm;
 	struct spa_bt_device *d;
 	struct spa_bt_transport *t = NULL;
@@ -1645,19 +1712,7 @@ static DBusHandlerResult profile_new_connection(DBusConnection *conn, DBusMessag
 	}
 
 	handler = dbus_message_get_path(m);
-#ifdef HAVE_BLUEZ_5_BACKEND_HSP_NATIVE
-	if (spa_streq(handler, PROFILE_HSP_AG))
-		profile = SPA_BT_PROFILE_HSP_HS;
-	else if (spa_streq(handler, PROFILE_HSP_HS))
-		profile = SPA_BT_PROFILE_HSP_AG;
-#endif
-#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
-	if (spa_streq(handler, PROFILE_HFP_AG))
-		profile = SPA_BT_PROFILE_HFP_HF;
-	else if (spa_streq(handler, PROFILE_HFP_HF))
-		profile = SPA_BT_PROFILE_HFP_AG;
-#endif
-
+	profile = path_to_profile(handler);
 	if (profile == SPA_BT_PROFILE_NULL) {
 		spa_log_warn(backend->log, "invalid handler %s", handler);
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -1667,7 +1722,7 @@ static DBusHandlerResult profile_new_connection(DBusConnection *conn, DBusMessag
 	dbus_message_iter_get_basic(&it[0], &path);
 
 	d = spa_bt_device_find(backend->monitor, path);
-	if (d == NULL) {
+	if (d == NULL || d->adapter == NULL) {
 		spa_log_warn(backend->log, "unknown device for path %s", path);
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	}
@@ -1784,19 +1839,7 @@ static DBusHandlerResult profile_request_disconnection(DBusConnection *conn, DBu
 	}
 
 	handler = dbus_message_get_path(m);
-#ifdef HAVE_BLUEZ_5_BACKEND_HSP_NATIVE
-	if (spa_streq(handler, PROFILE_HSP_AG))
-		profile = SPA_BT_PROFILE_HSP_HS;
-	else if (spa_streq(handler, PROFILE_HSP_HS))
-		profile = SPA_BT_PROFILE_HSP_AG;
-#endif
-#ifdef HAVE_BLUEZ_5_BACKEND_HFP_NATIVE
-	if (spa_streq(handler, PROFILE_HFP_AG))
-		profile = SPA_BT_PROFILE_HFP_HF;
-	else if (spa_streq(handler, PROFILE_HFP_HF))
-		profile = SPA_BT_PROFILE_HFP_AG;
-#endif
-
+	profile = path_to_profile(handler);
 	if (profile == SPA_BT_PROFILE_NULL) {
 		spa_log_warn(backend->log, "invalid handler %s", handler);
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -1806,7 +1849,7 @@ static DBusHandlerResult profile_request_disconnection(DBusConnection *conn, DBu
 	dbus_message_iter_get_basic(&it[0], &path);
 
 	d = spa_bt_device_find(backend->monitor, path);
-	if (d == NULL) {
+	if (d == NULL || d->adapter == NULL) {
 		spa_log_warn(backend->log, "unknown device for path %s", path);
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	}
