@@ -264,8 +264,7 @@ done:
 	return res;
 
 error:
-	if (client->core_resource)
-		pw_resource_errorf(client->core_resource, res, "client error %d (%s)",
+	pw_resource_errorf(client->core_resource, res, "client error %d (%s)",
 				res, spa_strerror(res));
 	goto done;
 }
@@ -291,7 +290,7 @@ client_busy_changed(void *data, bool busy)
 
 static void handle_client_error(struct pw_impl_client *client, int res)
 {
-	if (res == -EPIPE)
+	if (res == -EPIPE || res == -ECONNRESET)
 		pw_log_info("%p: client %p disconnected", client->protocol, client);
 	else
 		pw_log_error("%p: client %p error %d (%s)", client->protocol,
