@@ -29,6 +29,7 @@
 #include <sys/socket.h>
 
 #include <spa/utils/defs.h>
+#include <spa/utils/hook.h>
 #include <spa/utils/list.h>
 #include <pipewire/core.h>
 #include <pipewire/log.h>
@@ -48,6 +49,32 @@
 #include "server.h"
 #include "stream.h"
 
+#define client_emit_disconnect(c) spa_hook_list_call(&(c)->listener_list, struct client_events, disconnect, 0)
+
+struct client *client_new(struct server *server)
+{
+	struct client *client = calloc(1, sizeof(*client));
+	if (client == NULL)
+		return NULL;
+
+	client->ref = 1;
+	client->server = server;
+	client->impl = server->impl;
+	client->connect_tag = SPA_ID_INVALID;
+
+	pw_map_init(&client->streams, 16, 16);
+	spa_list_init(&client->out_messages);
+	spa_list_init(&client->operations);
+	spa_list_init(&client->pending_samples);
+	spa_list_init(&client->pending_streams);
+	spa_hook_list_init(&client->listener_list);
+
+	spa_list_append(&server->clients, &client->link);
+	server->n_clients++;
+
+	return client;
+}
+
 static int client_free_stream(void *item, void *data)
 {
 	struct stream *s = item;
@@ -72,6 +99,7 @@ bool client_detach(struct client *client)
 
 	/* remove from the `server->clients` list */
 	spa_list_remove(&client->link);
+	spa_list_append(&impl->cleanup_clients, &client->link);
 
 	server->n_clients--;
 	if (server->wait_clients > 0 && --server->wait_clients == 0) {
@@ -92,19 +120,24 @@ void client_disconnect(struct client *client)
 	if (client->disconnect)
 		return;
 
+	client_emit_disconnect(client);
+
 	/* the client must be detached from the server to disconnect */
 	spa_assert(client->server == NULL);
 
 	client->disconnect = true;
-	spa_list_append(&impl->cleanup_clients, &client->link);
 
 	pw_map_for_each(&client->streams, client_free_stream, client);
 
-	if (client->source)
+	if (client->source) {
 		pw_loop_destroy_source(impl->loop, client->source);
+		client->source = NULL;
+	}
 
-	if (client->manager)
+	if (client->manager) {
 		pw_manager_destroy(client->manager);
+		client->manager = NULL;
+	}
 }
 
 void client_free(struct client *client)
@@ -134,21 +167,20 @@ void client_free(struct client *client)
 	spa_list_consume(o, &client->operations, link)
 		operation_free(o);
 
-	if (client->core) {
-		client->disconnecting = true;
+	if (client->core)
 		pw_core_disconnect(client->core);
-	}
 
 	pw_map_clear(&client->streams);
 
+	pw_work_queue_cancel(impl->work_queue, client, SPA_ID_INVALID);
+
 	free(client->default_sink);
 	free(client->default_source);
 
-	if (client->props)
-		pw_properties_free(client->props);
+	pw_properties_free(client->props);
+	pw_properties_free(client->routes);
 
-	if (client->routes)
-		pw_properties_free(client->routes);
+	spa_hook_list_clean(&client->listener_list);
 
 	free(client);
 }
@@ -156,11 +188,16 @@ void client_free(struct client *client)
 int client_queue_message(struct client *client, struct message *msg)
 {
 	struct impl *impl = client->impl;
-	int res, mask;
+	int res;
 
 	if (msg == NULL)
 		return -EINVAL;
 
+	if (client->disconnect) {
+		res = -ENOTCONN;
+		goto error;
+	}
+
 	if (msg->length == 0) {
 		res = 0;
 		goto error;
@@ -172,13 +209,14 @@ int client_queue_message(struct client *client, struct message *msg)
 	msg->offset = 0;
 	spa_list_append(&client->out_messages, &msg->link);
 
-	mask = client->source->mask;
+	uint32_t mask = client->source->mask;
 	if (!SPA_FLAG_IS_SET(mask, SPA_IO_OUT)) {
-		client->need_flush = true;
 		SPA_FLAG_SET(mask, SPA_IO_OUT);
 		pw_loop_update_io(impl->loop, client->source, mask);
 	}
 
+	client->new_msg_since_last_flush = true;
+
 	return 0;
 
 error:
@@ -186,22 +224,20 @@ error:
 	return res;
 }
 
-int client_flush_messages(struct client *client)
+static int client_try_flush_messages(struct client *client)
 {
 	struct impl *impl = client->impl;
-	int res;
 
-	while (true) {
-		struct message *m;
+	pw_log_trace("client %p: flushing", client);
+
+	spa_assert(!client->disconnect);
+
+	while (!spa_list_is_empty(&client->out_messages)) {
+		struct message *m = spa_list_first(&client->out_messages, struct message, link);
 		struct descriptor desc;
-		void *data;
+		const void *data;
 		size_t size;
 
-		if (spa_list_is_empty(&client->out_messages))
-			break;
-
-		m = spa_list_first(&client->out_messages, struct message, link);
-
 		if (client->out_index < sizeof(desc)) {
 			desc.length = htonl(m->length);
 			desc.channel = htonl(m->channel);
@@ -224,18 +260,18 @@ int client_flush_messages(struct client *client)
 		}
 
 		while (true) {
-			res = send(client->source->fd, data, size, MSG_NOSIGNAL | MSG_DONTWAIT);
-			if (res < 0) {
-				res = -errno;
+			ssize_t sent = send(client->source->fd, data, size, MSG_NOSIGNAL | MSG_DONTWAIT);
+			if (sent < 0) {
+				int res = -errno;
 				if (res == -EINTR)
 					continue;
 				if (res != -EAGAIN && res != -EWOULDBLOCK)
-					pw_log_warn("client %p: send channel:%d %zu, error %d: %m",
+					pw_log_warn("client %p: send channel:%u %zu, error %d: %m",
 						    client, m->channel, size, res);
 				return res;
 			}
 
-			client->out_index += res;
+			client->out_index += sent;
 			break;
 		}
 	}
@@ -243,53 +279,116 @@ int client_flush_messages(struct client *client)
 	return 0;
 }
 
-int client_queue_subscribe_event(struct client *client, uint32_t mask, uint32_t event, uint32_t id)
+int client_flush_messages(struct client *client)
 {
-	struct impl *impl = client->impl;
-	struct message *reply, *m, *t;
+	client->new_msg_since_last_flush = false;
+
+	int res = client_try_flush_messages(client);
+	if (res >= 0) {
+		uint32_t mask = client->source->mask;
+
+		if (SPA_FLAG_IS_SET(mask, SPA_IO_OUT)) {
+			SPA_FLAG_CLEAR(mask, SPA_IO_OUT);
+			pw_loop_update_io(client->impl->loop, client->source, mask);
+		}
+	} else {
+		if (res != -EAGAIN && res != -EWOULDBLOCK)
+			return res;
+	}
+
+	return 0;
+}
+
+static bool drop_from_out_queue(struct client *client, struct message *m)
+{
+	spa_assert(!spa_list_is_empty(&client->out_messages));
+
+	struct message *first = spa_list_first(&client->out_messages, struct message, link);
+	if (m == first && client->out_index > 0)
+		return false;
+
+	message_free(client->impl, m, true, false);
+
+	return true;
+}
+
+/* returns true if an event with the (mask, event, index) triplet should be dropped because it is redundant */
+static bool client_prune_subscribe_events(struct client *client, uint32_t mask, uint32_t event, uint32_t index)
+{
+	struct message *m, *t;
+
+	if ((event & SUBSCRIPTION_EVENT_TYPE_MASK) == SUBSCRIPTION_EVENT_NEW)
+		return false;
+
+	/* NOTE: reverse iteration */
+	spa_list_for_each_safe_reverse(m, t, &client->out_messages, link) {
+		if (m->extra[0] != COMMAND_SUBSCRIBE_EVENT)
+			continue;
+		if ((m->extra[1] ^ event) & SUBSCRIPTION_EVENT_FACILITY_MASK)
+			continue;
+		if (m->extra[2] != index)
+			continue;
+
+		if ((event & SUBSCRIPTION_EVENT_TYPE_MASK) == SUBSCRIPTION_EVENT_REMOVE) {
+			/* This object is being removed, hence there is
+			 * point in keeping the old events regarding
+			 * entry in the queue. */
+
+			bool is_new = (m->extra[1] & SUBSCRIPTION_EVENT_TYPE_MASK) == SUBSCRIPTION_EVENT_NEW;
+
+			if (drop_from_out_queue(client, m)) {
+				pw_log_debug("client %p: dropped redundant event due to remove event for object %u",
+					     client, index);
+
+				/* if the NEW event for the current object could successfully be dropped,
+				   there is no need to deliver the REMOVE event */
+				if (is_new)
+					goto drop;
+			}
+
+			/* stop if the NEW event for the current object is reached */
+			if (is_new)
+				break;
+		}
+
+		if ((event & SUBSCRIPTION_EVENT_TYPE_MASK) == SUBSCRIPTION_EVENT_CHANGE) {
+			/* This object has changed. If a "new" or "change" event for
+			 * this object is still in the queue we can exit. */
+			goto drop;
+		}
+	}
+
+	return false;
+
+drop:
+	pw_log_debug("client %p: dropped redundant event for object %u", client, index);
+
+	return true;
+}
+
+int client_queue_subscribe_event(struct client *client, uint32_t mask, uint32_t event, uint32_t index)
+{
+	if (client->disconnect)
+		return -ENOTCONN;
 
 	if (!(client->subscribed & mask))
 		return 0;
 
-	pw_log_debug("client %p: SUBSCRIBE event:%08x id:%u", client, event, id);
+	pw_log_debug("client %p: SUBSCRIBE event:%08x index:%u", client, event, index);
 
-	if ((event & SUBSCRIPTION_EVENT_TYPE_MASK) != SUBSCRIPTION_EVENT_NEW) {
-		spa_list_for_each_safe_reverse(m, t, &client->out_messages, link) {
-			if (m->extra[0] != COMMAND_SUBSCRIBE_EVENT)
-				continue;
-			if ((m->extra[1] ^ event) & SUBSCRIPTION_EVENT_FACILITY_MASK)
-				continue;
-			if (m->extra[2] != id)
-				continue;
+	if (client_prune_subscribe_events(client, mask, event, index))
+		return 0;
 
-			if ((event & SUBSCRIPTION_EVENT_TYPE_MASK) == SUBSCRIPTION_EVENT_REMOVE) {
-				/* This object is being removed, hence there is
-				 * point in keeping the old events regarding
-				 * entry in the queue. */
-				message_free(impl, m, true, false);
-				pw_log_debug("client %p: dropped redundant event due to remove event", client);
-				continue;
-			}
-
-			if ((event & SUBSCRIPTION_EVENT_TYPE_MASK) == SUBSCRIPTION_EVENT_CHANGE) {
-				/* This object has changed. If a "new" or "change" event for
-				 * this object is still in the queue we can exit. */
-				pw_log_debug("client %p: dropped redundant event due to change event", client);
-				return 0;
-			}
-		}
-	}
-
-	reply = message_alloc(impl, -1, 0);
-	reply->extra[0] = COMMAND_SUBSCRIBE_EVENT,
-	reply->extra[1] = event,
-	reply->extra[2] = id,
+	struct message *reply = message_alloc(client->impl, -1, 0);
+	reply->extra[0] = COMMAND_SUBSCRIBE_EVENT;
+	reply->extra[1] = event;
+	reply->extra[2] = index;
 
 	message_put(reply,
 		TAG_U32, COMMAND_SUBSCRIBE_EVENT,
 		TAG_U32, -1,
 		TAG_U32, event,
-		TAG_U32, id,
+		TAG_U32, index,
 		TAG_INVALID);
 
 	return client_queue_message(client, reply);
