@@ -42,16 +42,19 @@
 #include "log.h"
 #include "module.h"
 
-static void on_module_unload(void *obj, void *data, int res, uint32_t id)
+static void on_module_unload(void *obj, void *data, int res, uint32_t index)
 {
 	struct module *module = obj;
-	module_unload(NULL, module);
+	module_unload(module);
 }
 
 void module_schedule_unload(struct module *module)
 {
-	struct impl *impl = module->impl;
-	pw_work_queue_add(impl->work_queue, module, 0, on_module_unload, impl);
+	if (module->unloading)
+		return;
+
+	pw_work_queue_add(module->impl->work_queue, module, 0, on_module_unload, NULL);
+	module->unloading = true;
 }
 
 struct module *module_new(struct impl *impl, const struct module_methods *methods, size_t user_data)
@@ -80,7 +83,7 @@ void module_add_listener(struct module *module,
 
 int module_load(struct client *client, struct module *module)
 {
-	pw_log_info("load module id:%u name:%s", module->idx, module->name);
+	pw_log_info("load module index:%u name:%s", module->index, module->name);
 	if (module->methods->load == NULL)
 		return -ENOTSUP;
 	/* subscription event is sent when the module does a
@@ -92,8 +95,10 @@ void module_free(struct module *module)
 {
 	struct impl *impl = module->impl;
 
-	if (module->idx != SPA_ID_INVALID)
-		pw_map_remove(&impl->modules, module->idx & INDEX_MASK);
+	module_emit_destroy(module);
+
+	if (module->index != SPA_ID_INVALID)
+		pw_map_remove(&impl->modules, module->index & MODULE_INDEX_MASK);
 
 	spa_hook_list_clean(&module->listener_list);
 	pw_work_queue_cancel(impl->work_queue, module, SPA_ID_INVALID);
@@ -105,7 +110,7 @@ void module_free(struct module *module)
 	free(module);
 }
 
-int module_unload(struct client *client, struct module *module)
+int module_unload(struct module *module)
 {
 	struct impl *impl = module->impl;
 	int res = 0;
@@ -113,16 +118,16 @@ int module_unload(struct client *client, struct module *module)
 	/* Note that client can be NULL (when the module is being unloaded
 	 * internally and not by a client request */
 
-	pw_log_info("unload module id:%u name:%s", module->idx, module->name);
+	pw_log_info("unload module index:%u name:%s", module->index, module->name);
 
 	if (module->methods->unload)
-		res = module->methods->unload(client, module);
+		res = module->methods->unload(module);
 
 	if (module->loaded)
 		broadcast_subscribe_event(impl,
 			SUBSCRIPTION_MASK_MODULE,
 			SUBSCRIPTION_EVENT_REMOVE | SUBSCRIPTION_EVENT_MODULE,
-			module->idx);
+			module->index);
 
 	module_free(module);
 
@@ -263,20 +268,20 @@ static const struct module_info module_list[] = {
 #ifdef HAVE_AVAHI
 	{ "module-zeroconf-publish", create_module_zeroconf_publish, },
 #endif
-#ifdef HAVE_ROC
 	{ "module-roc-sink", create_module_roc_sink, },
 	{ "module-roc-source", create_module_roc_source, },
-#endif
-	{ NULL, }
+	{ "module-x11-bell", create_module_x11_bell, },
 };
 
 static const struct module_info *find_module_info(const char *name)
 {
-	int i;
-	for (i = 0; module_list[i].name != NULL; i++) {
-		if (spa_streq(module_list[i].name, name))
-			return &module_list[i];
+	const struct module_info *info;
+
+	SPA_FOR_EACH_ELEMENT(module_list, info) {
+		if (spa_streq(info->name, name))
+			return info;
 	}
+
 	return NULL;
 }
 
@@ -295,13 +300,13 @@ struct module *module_create(struct client *client, const char *name, const char
 	if (module == NULL)
 		return NULL;
 
-	module->idx = pw_map_insert_new(&impl->modules, module);
-	if (module->idx == SPA_ID_INVALID) {
-		module_unload(client, module);
+	module->index = pw_map_insert_new(&impl->modules, module);
+	if (module->index == SPA_ID_INVALID) {
+		module_unload(module);
 		return NULL;
 	}
 	module->name = strdup(name);
 	module->args = args ? strdup(args) : NULL;
-	module->idx |= MODULE_FLAG;
+	module->index |= MODULE_FLAG;
 	return module;
 }
