@@ -32,6 +32,11 @@
 #include "../module.h"
 #include "registry.h"
 
+#define NAME "remap-sink"
+
+PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
+#define PW_LOG_TOPIC_DEFAULT mod_topic
+
 struct module_remap_source_data {
 	struct module *module;
 
@@ -62,8 +67,8 @@ static int module_remap_source_load(struct client *client, struct module *module
 	char *args;
 	size_t size;
 
-	pw_properties_setf(data->capture_props, PW_KEY_NODE_GROUP, "remap-source-%u", module->idx);
-	pw_properties_setf(data->playback_props, PW_KEY_NODE_GROUP, "remap-source-%u", module->idx);
+	pw_properties_setf(data->capture_props, PW_KEY_NODE_GROUP, "remap-source-%u", module->index);
+	pw_properties_setf(data->playback_props, PW_KEY_NODE_GROUP, "remap-source-%u", module->index);
 
 	f = open_memstream(&args, &size);
 	fprintf(f, "{");
@@ -90,7 +95,7 @@ static int module_remap_source_load(struct client *client, struct module *module
 	return 0;
 }
 
-static int module_remap_source_unload(struct client *client, struct module *module)
+static int module_remap_source_unload(struct module *module)
 {
 	struct module_remap_source_data *d = module->user_data;
 
@@ -134,9 +139,9 @@ static void position_to_props(struct spa_audio_info_raw *info, struct pw_propert
 	uint32_t i;
 
 	pw_properties_setf(props, SPA_KEY_AUDIO_CHANNELS, "%u", info->channels);
-	p = s = alloca(info->channels * 6);
+	p = s = alloca(info->channels * 8);
 	for (i = 0; i < info->channels; i++)
-		p += snprintf(p, 6, "%s%s", i == 0 ? "" : ",",
+		p += spa_scnprintf(p, 8, "%s%s", i == 0 ? "" : ",",
 				channel_id2name(info->position[i]));
 	pw_properties_set(props, SPA_KEY_AUDIO_POSITION, s);
 }
@@ -146,11 +151,13 @@ struct module *create_module_remap_source(struct impl *impl, const char *argumen
 	struct module *module;
 	struct module_remap_source_data *d;
 	struct pw_properties *props = NULL, *playback_props = NULL, *capture_props = NULL;
-	const char *str;
+	const char *str, *master;
 	struct spa_audio_info_raw capture_info = { 0 };
 	struct spa_audio_info_raw playback_info = { 0 };
 	int res;
 
+	PW_LOG_TOPIC_INIT(mod_topic);
+
 	props = pw_properties_new_dict(&SPA_DICT_INIT_ARRAY(module_remap_source_info));
 	capture_props = pw_properties_new(NULL, NULL);
 	playback_props = pw_properties_new(NULL, NULL);
@@ -161,16 +168,40 @@ struct module *create_module_remap_source(struct impl *impl, const char *argumen
 	if (argument)
 		module_args_add_props(props, argument);
 
+	master = pw_properties_get(props, "master");
+	if (pw_properties_get(props, "source_name") == NULL) {
+		pw_properties_setf(props, "source_name", "%s.remapped",
+				master ? master : "default");
+	}
 	if ((str = pw_properties_get(props, "source_name")) != NULL) {
-		pw_properties_set(props, PW_KEY_NODE_NAME, str);
+		pw_properties_set(playback_props, PW_KEY_NODE_NAME, str);
+		pw_properties_setf(capture_props, PW_KEY_NODE_NAME, "input.%s", str);
 		pw_properties_set(props, "source_name", NULL);
 	}
 	if ((str = pw_properties_get(props, "source_properties")) != NULL) {
 		module_args_add_props(playback_props, str);
 		pw_properties_set(props, "source_properties", NULL);
 	}
-	pw_properties_set(playback_props, PW_KEY_MEDIA_CLASS, "Audio/Source");
+	if (pw_properties_get(playback_props, PW_KEY_MEDIA_CLASS) == NULL)
+		pw_properties_set(playback_props, PW_KEY_MEDIA_CLASS, "Audio/Source");
+	if (pw_properties_get(playback_props, PW_KEY_DEVICE_CLASS) == NULL)
+		pw_properties_set(playback_props, PW_KEY_DEVICE_CLASS, "filter");
 
+	if ((str = pw_properties_get(playback_props, PW_KEY_MEDIA_NAME)) != NULL)
+		pw_properties_set(props, PW_KEY_MEDIA_NAME, str);
+	if ((str = pw_properties_get(playback_props, PW_KEY_NODE_DESCRIPTION)) != NULL) {
+		pw_properties_set(props, PW_KEY_NODE_DESCRIPTION, str);
+	} else {
+		str = pw_properties_get(playback_props, PW_KEY_NODE_NAME);
+		if (master != NULL || str == NULL) {
+			pw_properties_setf(props, PW_KEY_NODE_DESCRIPTION,
+					"Remapped %s source",
+					master ? master : "default");
+		} else {
+			pw_properties_setf(props, PW_KEY_NODE_DESCRIPTION,
+					"%s source", str);
+		}
+	}
 	if ((str = pw_properties_get(props, "master")) != NULL) {
 		pw_properties_set(capture_props, PW_KEY_NODE_TARGET, str);
 		pw_properties_set(props, "master", NULL);
@@ -200,7 +231,7 @@ struct module *create_module_remap_source(struct impl *impl, const char *argumen
 	if ((str = pw_properties_get(props, "remix")) != NULL) {
 		/* Note that the boolean is inverted */
 		pw_properties_set(capture_props, PW_KEY_STREAM_DONT_REMIX,
-				pw_properties_parse_bool(str) ? "false" : "true");
+				module_args_parse_bool(str) ? "false" : "true");
 		pw_properties_set(props, "remix", NULL);
 	}
 
