@@ -134,7 +134,8 @@ static int context_set_freewheel(struct pw_context *context, bool freewheel)
 		res = pw_thread_utils_drop_rt(thr);
 	} else {
 		pw_log_info("%p: exit freewheel", context);
-		res = pw_thread_utils_acquire_rt(thr, 88);
+		// Use the priority as configured within the realtime module
+		res = pw_thread_utils_acquire_rt(thr, -1);
 	}
 	if (res < 0)
 		pw_log_info("%p: freewheel error:%s", context, spa_strerror(res));
@@ -272,10 +273,8 @@ struct pw_context *pw_context_new(struct pw_loop *main_loop,
 	n_support = pw_get_support(this->support, SPA_N_ELEMENTS(this->support) - 6);
 	cpu = spa_support_find(this->support, n_support, SPA_TYPE_INTERFACE_CPU);
 
-	if ((str = pw_properties_get(conf, "context.properties")) != NULL) {
-		pw_properties_update_string(properties, str, strlen(str));
-		pw_log_info("%p: parsed context.properties section", this);
-	}
+	res = pw_context_conf_update_props(this, "context.properties", properties);
+	pw_log_info("%p: parsed %d context.properties items", this, res);
 
 	if ((str = getenv("PIPEWIRE_CORE"))) {
 		pw_log_info("using core.name from environment: %s", str);
@@ -893,24 +892,31 @@ static int ensure_state(struct pw_impl_node *node, bool running)
 	return pw_impl_node_set_state(node, state);
 }
 
-static int collect_nodes(struct pw_context *context, struct pw_impl_node *driver)
+static int collect_nodes(struct pw_context *context, struct pw_impl_node *node)
 {
 	struct spa_list queue;
-	struct pw_impl_node *n, *t;
+	struct pw_impl_node *n, *t, *driver;
 	struct pw_impl_port *p;
 	struct pw_impl_link *l;
 
-	spa_list_consume(t, &driver->follower_list, follower_link) {
-		spa_list_remove(&t->follower_link);
-		spa_list_init(&t->follower_link);
+	pw_log_debug("node %p: '%s'", node, node->name);
+
+	if (node->driver) {
+		driver = node;
+		spa_list_consume(t, &driver->follower_list, follower_link) {
+			spa_list_remove(&t->follower_link);
+			spa_list_init(&t->follower_link);
+		}
+	} else {
+		driver = node->driver_node;
+		if (driver == NULL)
+			return -EINVAL;
 	}
 
-	pw_log_debug("driver %p: '%s'", driver, driver->name);
-
-	/* start with driver in the queue */
+	/* start with node in the queue */
 	spa_list_init(&queue);
-	spa_list_append(&queue, &driver->sort_link);
-	driver->visited = true;
+	spa_list_append(&queue, &node->sort_link);
+	node->visited = true;
 
 	/* now follow all the links from the nodes in the queue
 	 * and add the peers to the queue. */
@@ -980,26 +986,32 @@ static int collect_nodes(struct pw_context *context, struct pw_impl_node *driver
 }
 
 static inline void get_quantums(struct pw_context *context, uint32_t *def,
-		uint32_t *min, uint32_t *max, uint32_t *rate)
+		uint32_t *min, uint32_t *max, uint32_t *limit, uint32_t *rate)
 {
 	struct settings *s = &context->settings;
-	*def = s->clock_force_quantum == 0 ? s->clock_quantum : s->clock_force_quantum;
-	*min = s->clock_force_quantum == 0 ? s->clock_min_quantum : s->clock_force_quantum;
-	*max = s->clock_force_quantum == 0 ? s->clock_max_quantum : s->clock_force_quantum;
-	*rate = s->clock_force_quantum == 0 ? s->clock_rate : s->clock_force_rate;
+	if (s->clock_force_quantum != 0) {
+		*def = *min = *max = s->clock_force_quantum;
+		*rate = 0;
+	} else {
+		*def = s->clock_quantum;
+		*min = s->clock_min_quantum;
+		*max = s->clock_max_quantum;
+		*rate = s->clock_rate;
+	}
+	*limit = s->clock_quantum_limit;
 }
 
 static inline uint32_t *get_rates(struct pw_context *context, uint32_t *def, uint32_t *n_rates,
-		bool *force_rate)
+		bool *force)
 {
 	struct settings *s = &context->settings;
 	if (s->clock_force_rate != 0) {
-		*force_rate = true;
+		*force = true;
 		*n_rates = 1;
 		*def = s->clock_force_rate;
 		return &s->clock_force_rate;
 	} else {
-		*force_rate = false;
+		*force = false;
 		*n_rates = s->n_clock_rates;
 		*def = s->clock_rate;
 		return s->clock_rates;
@@ -1052,10 +1064,11 @@ static bool rates_contains(uint32_t *rates, uint32_t n_rates, uint32_t rate)
 int pw_context_recalc_graph(struct pw_context *context, const char *reason)
 {
 	struct impl *impl = SPA_CONTAINER_OF(context, struct impl, this);
+	struct settings *settings = &context->settings;
 	struct pw_impl_node *n, *s, *target, *fallback;
-	uint32_t max_quantum, min_quantum, def_quantum, rate_quantum;
+	uint32_t max_quantum, min_quantum, def_quantum, lim_quantum, rate_quantum;
 	uint32_t *rates, n_rates, def_rate;
-	bool freewheel = false, force_rate;
+	bool freewheel = false, global_force_rate, force_rate, global_force_quantum;
 
 	pw_log_info("%p: busy:%d reason:%s", context, impl->recalc, reason);
 
@@ -1067,8 +1080,11 @@ int pw_context_recalc_graph(struct pw_context *context, const char *reason)
 again:
 	impl->recalc = true;
 
-	get_quantums(context, &def_quantum, &min_quantum, &max_quantum, &rate_quantum);
-	rates = get_rates(context, &def_rate, &n_rates, &force_rate);
+	get_quantums(context, &def_quantum, &min_quantum, &max_quantum, &lim_quantum, &rate_quantum);
+	rates = get_rates(context, &def_rate, &n_rates, &global_force_rate);
+
+	global_force_quantum = rate_quantum == 0;
+	force_rate = global_force_rate;
 
 	/* start from all drivers and group all nodes that are linked
 	 * to it. Some nodes are not (yet) linked to anything and they
@@ -1135,8 +1151,10 @@ again:
 			pw_impl_node_set_driver(n, t);
 			if (t == NULL)
 				ensure_state(n, false);
-			else
+			else {
 				t->passive = false;
+				collect_nodes(context, n);
+			}
 		}
 		n->visited = false;
 	}
@@ -1148,6 +1166,7 @@ again:
 		struct spa_fraction max_latency = SPA_FRACTION(0, 0);
 		struct spa_fraction rate = SPA_FRACTION(0, 0);
 		uint32_t quantum, target_rate, current_rate;
+		uint64_t quantum_stamp = 0, rate_stamp = 0;
 
 		if (!n->driving || n->exported)
 			continue;
@@ -1159,6 +1178,20 @@ again:
 				lock_quantum |= s->lock_quantum;
 				lock_rate |= s->lock_rate;
 			}
+			if (!global_force_quantum && s->force_quantum > 0 &&
+			    s->stamp > quantum_stamp) {
+				def_quantum = min_quantum = max_quantum = s->force_quantum;
+				rate_quantum = 0;
+				quantum_stamp = s->stamp;
+			}
+			if (!global_force_rate && s->force_rate > 0 &&
+			    s->stamp > rate_stamp) {
+				def_rate = s->force_rate;
+				force_rate = true;
+				n_rates = 1;
+				rates = &s->force_rate;
+				rate_stamp = s->stamp;
+			}
 
 			/* smallest latencies */
 			if (latency.denom == 0 ||
@@ -1180,21 +1213,22 @@ again:
 				running = !n->passive;
 		}
 
-		/* calculate desired rate */
-		target_rate = def_rate;
-		if (rate.denom != 0 && rate.num == 1) {
-			if (rates_contains(rates, n_rates, rate.denom))
-				target_rate = rate.denom;
-		}
 		if (force_rate)
 			lock_rate = false;
 
 		current_rate = n->current_rate.denom;
-		if (target_rate != current_rate && lock_rate)
-			target_rate = current_rate;
-		else if (target_rate != current_rate && !force_rate &&
-		    (n->info.state > PW_NODE_STATE_IDLE))
+		if (lock_rate ||
+		    (!force_rate &&
+		    (n->info.state > PW_NODE_STATE_IDLE)))
 			target_rate = current_rate;
+		else {
+			/* calculate desired rate */
+			target_rate = def_rate;
+			if (rate.denom != 0 && rate.num == 1) {
+				if (rates_contains(rates, n_rates, rate.denom))
+					target_rate = rate.denom;
+			}
+		}
 
 		if (target_rate != current_rate) {
 			pw_log_info("(%s-%u) state:%s new rate:%u->%u",
@@ -1204,7 +1238,7 @@ again:
 					target_rate);
 
 			if (force_rate) {
-				if (context->settings.clock_rate_update_mode == CLOCK_RATE_UPDATE_MODE_HARD)
+				if (settings->clock_rate_update_mode == CLOCK_RATE_UPDATE_MODE_HARD)
 					suspend_driver(context, n);
 			} else {
 				if (n->info.state >= PW_NODE_STATE_IDLE)
@@ -1220,6 +1254,7 @@ again:
 		if (rate_quantum != 0 && current_rate != rate_quantum) {
 			def_quantum = def_quantum * current_rate / rate_quantum;
 			min_quantum = min_quantum * current_rate / rate_quantum;
+			max_quantum = max_quantum * current_rate / rate_quantum;
 		}
 
 		/* calculate desired quantum */
@@ -1233,8 +1268,9 @@ again:
 		if (latency.denom != 0)
 			quantum = (latency.num * current_rate / latency.denom);
 		quantum = SPA_CLAMP(quantum, min_quantum, max_quantum);
+		quantum = SPA_MIN(quantum, lim_quantum);
 
-		if (context->settings.clock_power_of_two_quantum)
+		if (settings->clock_power_of_two_quantum)
 			quantum = flp2(quantum);
 
 		if (running && quantum != n->current_quantum && !lock_quantum) {
@@ -1246,6 +1282,12 @@ again:
 			n->current_pending = true;
 		}
 
+		if (n->info.state < PW_NODE_STATE_RUNNING && n->current_pending) {
+			n->rt.position->clock.duration = n->current_quantum;
+			n->rt.position->clock.rate = n->current_rate;
+			n->current_pending = false;
+		}
+
 		pw_log_debug("%p: driving %p running:%d passive:%d quantum:%u '%s'",
 				context, n, running, n->passive, quantum, n->name);
 
