@@ -32,7 +32,7 @@ extern "C" {
 #include <spa/utils/defs.h>
 #include <spa/utils/hook.h>
 
-/** \defgroup pw_context PipeWire Context
+/** \defgroup pw_context Context
  *
  * \brief The PipeWire context object manages all locally available
  * resources. It is used by both clients and servers.
@@ -108,8 +108,21 @@ const struct pw_properties *pw_context_get_properties(struct pw_context *context
 /** Update the context properties */
 int pw_context_update_properties(struct pw_context *context, const struct spa_dict *dict);
 
-/** Get a config section for this context. Since 0.3.22 */
+/** Get a config section for this context. Since 0.3.22, deprecated,
+ * use pw_context_conf_section_for_each(). */
 const char *pw_context_get_conf_section(struct pw_context *context, const char *section);
+/** Parse a standard config section for this context. Since 0.3.22 */
+int pw_context_parse_conf_section(struct pw_context *context,
+		struct pw_properties *conf, const char *section);
+
+/** update properties from a section into props. Since 0.3.45 */
+int pw_context_conf_update_props(struct pw_context *context, const char *section,
+		struct pw_properties *props);
+/** emit callback for all config sections. Since 0.3.45 */
+int pw_context_conf_section_for_each(struct pw_context *context, const char *section,
+		int (*callback) (void *data, const char *location, const char *section,
+			const char *str, size_t len),
+		void *data);
 
 /** Get the context support objects */
 const struct spa_support *pw_context_get_support(struct pw_context *context, uint32_t *n_support);
