@@ -168,6 +168,7 @@ struct stream {
 	unsigned int process_rt:1;
 	unsigned int driving:1;
 	unsigned int using_trigger:1;
+	unsigned int trigger:1;
 };
 
 static int get_param_index(uint32_t id)
@@ -1341,8 +1342,7 @@ stream_new(struct pw_context *context, const char *name,
 		res = -errno;
 		goto error_properties;
 	}
-	if ((str = pw_context_get_conf_section(context, "stream.properties")) != NULL)
-		pw_properties_update_string(props, str, strlen(str));
+	pw_context_conf_update_props(context, "stream.properties", props);
 
 	if (pw_properties_get(props, PW_KEY_STREAM_IS_LIVE) == NULL)
 		pw_properties_set(props, PW_KEY_STREAM_IS_LIVE, "true");
@@ -1357,6 +1357,17 @@ stream_new(struct pw_context *context, const char *name,
 	}
 	if ((str = getenv("PIPEWIRE_LATENCY")) != NULL)
 		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
+	if ((str = getenv("PIPEWIRE_RATE")) != NULL)
+		pw_properties_set(props, PW_KEY_NODE_RATE, str);
+	if ((str = getenv("PIPEWIRE_QUANTUM")) != NULL) {
+		struct spa_fraction q;
+		if (sscanf(str, "%u/%u", &q.num, &q.denom) == 2 && q.denom != 0) {
+			pw_properties_setf(props, PW_KEY_NODE_RATE,
+					"1/%u", q.denom);
+			pw_properties_setf(props, PW_KEY_NODE_LATENCY,
+					"%u/%u", q.num, q.denom);
+		}
+	}
 
 	spa_hook_list_init(&impl->hooks);
 	this->properties = props;
@@ -1766,6 +1777,10 @@ pw_stream_connect(struct pw_stream *stream,
 		pw_properties_set(stream->properties, PW_KEY_NODE_EXCLUSIVE, "true");
 	if (flags & PW_STREAM_FLAG_DONT_RECONNECT)
 		pw_properties_set(stream->properties, PW_KEY_NODE_DONT_RECONNECT, "true");
+	if (flags & PW_STREAM_FLAG_TRIGGER) {
+		pw_properties_set(stream->properties, PW_KEY_NODE_TRIGGER, "true");
+		impl->trigger = true;
+	}
 
 	if ((str = pw_properties_get(stream->properties, "mem.warn-mlock")) != NULL)
 		impl->warn_mlock = pw_properties_parse_bool(str);
@@ -2215,7 +2230,7 @@ int pw_stream_trigger_process(struct pw_stream *stream)
 	/* flag to check for old or new behaviour */
 	impl->using_trigger = true;
 
-	if (!impl->driving) {
+	if (!impl->driving && !impl->trigger) {
 		res = trigger_request_process(impl);
 	} else {
 		if (impl->direction == SPA_DIRECTION_OUTPUT &&
