@@ -22,6 +22,7 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
+#include <string.h>
 #include <unistd.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -29,24 +30,32 @@
 #include <getopt.h>
 #include <limits.h>
 #include <math.h>
+#include <fnmatch.h>
+
+#if !defined(FNM_EXTMATCH)
+#define FNM_EXTMATCH 0
+#endif
 
 #include <spa/utils/result.h>
+#include <spa/utils/string.h>
 #include <spa/pod/iter.h>
 #include <spa/debug/types.h>
 #include <spa/utils/json.h>
+#include <spa/utils/ansi.h>
+#include <spa/utils/string.h>
 
 #include <pipewire/pipewire.h>
-#include <extensions/metadata.h>
+#include <pipewire/extensions/metadata.h>
 
 #define INDENT 2
 
 static bool colors = false;
 
-#define NORMAL	(colors ? "\x1B[0m":"")
-#define LITERAL	(colors ? "\x1B[95m":"")
-#define NUMBER	(colors ? "\x1B[96m":"")
-#define STRING	(colors ? "\x1B[92m":"")
-#define KEY	(colors ? "\x1B[94m":"")
+#define NORMAL	(colors ? SPA_ANSI_RESET : "")
+#define LITERAL	(colors ? SPA_ANSI_BRIGHT_MAGENTA : "")
+#define NUMBER	(colors ? SPA_ANSI_BRIGHT_CYAN : "")
+#define STRING	(colors ? SPA_ANSI_BRIGHT_GREEN : "")
+#define KEY	(colors ? SPA_ANSI_BRIGHT_BLUE : "")
 
 struct data {
 	struct pw_main_loop *loop;
@@ -62,7 +71,7 @@ struct data {
 
 	struct spa_list object_list;
 
-	uint32_t id;
+	const char *pattern;
 
 	FILE *out;
 	int level;
@@ -90,6 +99,7 @@ struct class {
 	const void *events;
 	void (*destroy) (struct object *object);
 	void (*dump) (struct object *object);
+	const char *name_key;
 };
 
 struct object {
@@ -155,7 +165,7 @@ static struct param *add_param(struct spa_list *params, uint32_t id, const struc
 
 	p->id = id;
 	if (param != NULL) {
-		p->param = SPA_MEMBER(p, sizeof(*p), struct spa_pod);
+		p->param = SPA_PTROFF(p, sizeof(*p), struct spa_pod);
 		memcpy(p->param, param, SPA_POD_SIZE(param));
 	} else {
 		clear_params(params, id);
@@ -196,8 +206,7 @@ static void object_destroy(struct object *o)
 	spa_list_remove(&o->link);
 	if (o->proxy)
 		pw_proxy_destroy(o->proxy);
-	if (o->props)
-		pw_properties_free(o->props);
+	pw_properties_free(o->props);
 	clear_params(&o->param_list, SPA_ID_INVALID);
 	clear_params(&o->pending_list, SPA_ID_INVALID);
 	free(o->type);
@@ -245,12 +254,16 @@ static void put_end(struct data *d, const char *type, uint32_t flags)
 	d->state = (d->state & STATE_MASK) + STATE_COMMA - (flags & STATE_SIMPLE);
 }
 
+static void put_encoded_string(struct data *d, const char *key, const char *val)
+{
+	put_fmt(d, key, "%s%s%s", STRING, val, NORMAL);
+}
 static void put_string(struct data *d, const char *key, const char *val)
 {
 	int size = (strlen(val) + 1) * 4;
 	char *str = alloca(size);
 	spa_json_encode_string(str, size, val);
-	put_fmt(d, key, "%s%s%s", STRING, str, NORMAL);
+	put_encoded_string(d, key, str);
 }
 
 static void put_literal(struct data *d, const char *key, const char *val)
@@ -270,20 +283,16 @@ static void put_double(struct data *d, const char *key, double val)
 
 static void put_value(struct data *d, const char *key, const char *val)
 {
-	char *end;
-	long int li;
+	int64_t li;
 	double dv;
 
 	if (val == NULL)
 		put_literal(d, key, "null");
-	else if (strcmp(val, "true") == 0 ||
-	    strcmp(val, "false") == 0)
+	else if (spa_streq(val, "true") || spa_streq(val, "false"))
 		put_literal(d, key, val);
-	else if ((li = strtol(val, &end, 10)) != LONG_MIN &&
-	    errno != -ERANGE && *end == '\0')
+	else if (spa_atoi64(val, &li, 10))
 		put_int(d, key, li);
-	else if ((dv = strtod(val, &end)) != HUGE_VAL &&
-	    errno != -ERANGE && *end == '\0')
+	else if (spa_atod(val, &dv))
 		put_double(d, key, dv);
 	else
 		put_string(d, key, val);
@@ -375,14 +384,16 @@ static void put_pod_value(struct data *d, const char *key, const struct spa_type
 					SPA_POD_CONTENTS(struct spa_pod, &b->child),
 					b->child.size);
 		} else {
-			void *p;
-			static const char *range_labels[] = { "default", "min", "max", NULL };
-			static const char *step_labels[] = { "default", "min", "max", "step", NULL };
-			static const char *enum_labels[] = { "default", "alt%u" };
-			static const char *flags_labels[] = { "default", "flag%u" };
-			const char **labels, *label;
+			static const char * const range_labels[] = { "default", "min", "max", NULL };
+			static const char * const step_labels[] = { "default", "min", "max", "step", NULL };
+			static const char * const enum_labels[] = { "default", "alt%u" };
+			static const char * const flags_labels[] = { "default", "flag%u" };
+
+			const char * const *labels;
+			const char *label;
 			char buffer[64];
 			int max_labels, flags = 0;
+			void *p;
 
 			switch (b->type) {
 			case SPA_CHOICE_Range:
@@ -436,9 +447,16 @@ static void put_pod_value(struct data *d, const char *key, const struct spa_type
 		info = ti ? ti->values : info;
 
 		SPA_POD_OBJECT_BODY_FOREACH(b, size, p) {
+			char fallback[32];
+			const char *name;
+
 			ii = spa_debug_type_find(info, p->key);
-			put_pod_value(d,
-					ii ? spa_debug_type_short_name(ii->name) : "unknown",
+			name = ii ? spa_debug_type_short_name(ii->name) : NULL;
+			if (name == NULL) {
+				snprintf(fallback, sizeof(fallback), "id-%08x", p->key);
+				name = fallback;
+			}
+			put_pod_value(d, name,
 					ii ? ii->values : NULL,
 					p->value.type,
 					SPA_POD_CONTENTS(struct spa_pod_prop, p),
@@ -504,7 +522,7 @@ struct flags_info {
 };
 
 static void put_flags(struct data *d, const char *key,
-		uint64_t flags, struct flags_info *info)
+		uint64_t flags, const struct flags_info *info)
 {
 	uint32_t i;
 	put_begin(d, key, "[", STATE_SIMPLE);
@@ -518,12 +536,14 @@ static void put_flags(struct data *d, const char *key,
 /* core */
 static void core_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_core_info *i = d->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "props", PW_CORE_CHANGE_MASK_PROPS },
 		{ NULL, 0 },
 	};
+
+	struct data *d = o->data;
+	struct pw_core_info *i = d->info;
+
 	put_begin(d, "info", "{", 0);
 	put_int(d, "cookie", i->cookie);
 	put_value(d, "user-name", i->user_name);
@@ -539,17 +559,20 @@ static const struct class core_class = {
 	.type = PW_TYPE_INTERFACE_Core,
 	.version = PW_VERSION_CORE,
 	.dump = core_dump,
+	.name_key = PW_KEY_CORE_NAME,
 };
 
 /* client */
 static void client_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_client_info *i = o->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "props", PW_CLIENT_CHANGE_MASK_PROPS },
 		{ NULL, 0 },
 	};
+
+	struct data *d = o->data;
+	struct pw_client_info *i = o->info;
+
 	put_begin(d, "info", "{", 0);
 	put_flags(d, "change-mask", i->change_mask, fl);
 	put_dict(d, "props", i->props);
@@ -593,17 +616,20 @@ static const struct class client_class = {
 	.events = &client_events,
 	.destroy = client_destroy,
 	.dump = client_dump,
+	.name_key = PW_KEY_APP_NAME,
 };
 
 /* module */
 static void module_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_module_info *i = o->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "props", PW_MODULE_CHANGE_MASK_PROPS },
 		{ NULL, 0 },
 	};
+
+	struct data *d = o->data;
+	struct pw_module_info *i = o->info;
+
 	put_begin(d, "info", "{", 0);
 	put_value(d, "name", i->name);
 	put_value(d, "filename", i->filename);
@@ -650,17 +676,20 @@ static const struct class module_class = {
 	.events = &module_events,
 	.destroy = module_destroy,
 	.dump = module_dump,
+	.name_key = PW_KEY_MODULE_NAME,
 };
 
 /* factory */
 static void factory_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_factory_info *i = o->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "props", PW_FACTORY_CHANGE_MASK_PROPS },
 		{ NULL, 0 },
 	};
+
+	struct data *d = o->data;
+	struct pw_factory_info *i = o->info;
+
 	put_begin(d, "info", "{", 0);
 	put_value(d, "name", i->name);
 	put_value(d, "type", i->type);
@@ -707,18 +736,21 @@ static const struct class factory_class = {
 	.events = &factory_events,
 	.destroy = factory_destroy,
 	.dump = factory_dump,
+	.name_key = PW_KEY_FACTORY_NAME,
 };
 
 /* device */
 static void device_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_device_info *i = o->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "props", PW_DEVICE_CHANGE_MASK_PROPS },
 		{ "params", PW_DEVICE_CHANGE_MASK_PARAMS },
 		{ NULL, 0 },
 	};
+
+	struct data *d = o->data;
+	struct pw_device_info *i = o->info;
+
 	put_begin(d, "info", "{", 0);
 	put_flags(d, "change-mask", i->change_mask, fl);
 	put_dict(d, "props", i->props);
@@ -789,14 +821,13 @@ static const struct class device_class = {
 	.events = &device_events,
 	.destroy = device_destroy,
 	.dump = device_dump,
+	.name_key = PW_KEY_DEVICE_NAME,
 };
 
 /* node */
 static void node_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_node_info *i = o->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "input-ports", PW_NODE_CHANGE_MASK_INPUT_PORTS },
 		{ "output-ports", PW_NODE_CHANGE_MASK_OUTPUT_PORTS },
 		{ "state", PW_NODE_CHANGE_MASK_STATE },
@@ -804,6 +835,10 @@ static void node_dump(struct object *o)
 		{ "params", PW_NODE_CHANGE_MASK_PARAMS },
 		{ NULL, 0 },
 	};
+
+	struct data *d = o->data;
+	struct pw_node_info *i = o->info;
+
 	put_begin(d, "info", "{", 0);
 	put_int(d, "max-input-ports", i->max_input_ports);
 	put_int(d, "max-output-ports", i->max_output_ports);
@@ -883,18 +918,21 @@ static const struct class node_class = {
 	.events = &node_events,
 	.destroy = node_destroy,
 	.dump = node_dump,
+	.name_key = PW_KEY_NODE_NAME,
 };
 
 /* port */
 static void port_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_port_info *i = o->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "props", PW_PORT_CHANGE_MASK_PROPS },
 		{ "params", PW_PORT_CHANGE_MASK_PARAMS },
 		{ NULL, },
 	};
+
+	struct data *d = o->data;
+	struct pw_port_info *i = o->info;
+
 	put_begin(d, "info", "{", 0);
 	put_value(d, "direction", pw_direction_as_string(i->direction));
 	put_flags(d, "change-mask", i->change_mask, fl);
@@ -966,19 +1004,22 @@ static const struct class port_class = {
 	.events = &port_events,
 	.destroy = port_destroy,
 	.dump = port_dump,
+	.name_key = PW_KEY_PORT_NAME,
 };
 
 /* link */
 static void link_dump(struct object *o)
 {
-	struct data *d = o->data;
-	struct pw_link_info *i = o->info;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "state", PW_LINK_CHANGE_MASK_STATE },
 		{ "format", PW_LINK_CHANGE_MASK_FORMAT },
 		{ "props", PW_LINK_CHANGE_MASK_PROPS },
 		{ NULL, },
 	};
+
+	struct data *d = o->data;
+	struct pw_link_info *i = o->info;
+
 	put_begin(d, "info", "{", 0);
 	put_int(d, "output-node-id", i->output_node_id);
 	put_int(d, "output-port-id", i->output_port_id);
@@ -1037,6 +1078,43 @@ static const struct class link_class = {
 	.dump = link_dump,
 };
 
+static void json_dump_val(struct data *d, const char *key, struct spa_json *it, const char *value, int len)
+{
+	struct spa_json sub;
+	if (spa_json_is_array(value, len)) {
+		put_begin(d, key, "[", STATE_SIMPLE);
+		spa_json_enter(it, &sub);
+		while ((len = spa_json_next(&sub, &value)) > 0) {
+			json_dump_val(d, NULL, &sub, value, len);
+		}
+		put_end(d, "]", STATE_SIMPLE);
+	} else if (spa_json_is_object(value, len)) {
+		char val[1024];
+		put_begin(d, key, "{", STATE_SIMPLE);
+		spa_json_enter(it, &sub);
+		while (spa_json_get_string(&sub, val, sizeof(val)) > 0) {
+			if ((len = spa_json_next(&sub, &value)) <= 0)
+				break;
+			json_dump_val(d, val, &sub, value, len);
+		}
+		put_end(d, "}", STATE_SIMPLE);
+	} else if (spa_json_is_string(value, len)) {
+		put_encoded_string(d, key, strndupa(value, len));
+	} else {
+		put_value(d, key, strndupa(value, len));
+	}
+}
+
+static void json_dump(struct data *d, const char *key, const char *value)
+{
+	struct spa_json it[1];
+	int len;
+	const char *val;
+	spa_json_init(&it[0], value, strlen(value));
+	if ((len = spa_json_next(&it[0], &val)) >= 0)
+		json_dump_val(d, key, &it[0], val, len);
+}
+
 /* metadata */
 
 struct metadata_entry {
@@ -1061,7 +1139,10 @@ static void metadata_dump(struct object *o)
 		put_int(d, "subject", e->subject);
 		put_value(d, "key", e->key);
 		put_value(d, "type", e->type);
-		put_value(d, "value", e->value);
+		if (e->type != NULL && spa_streq(e->type, "Spa:String:JSON"))
+			json_dump(d, "value", e->value);
+		else
+			put_value(d, "value", e->value);
 		put_end(d, "}", STATE_SIMPLE);
 		e->changed = 0;
 	}
@@ -1073,7 +1154,7 @@ static struct metadata_entry *metadata_find(struct object *o, uint32_t subject,
 	struct metadata_entry *e;
 	spa_list_for_each(e, &o->data_list, link) {
 		if ((e->subject == subject) &&
-		    (key == NULL || strcmp(e->key, key) == 0))
+		    (key == NULL || spa_streq(e->key, key)))
 			return e;
 	}
 	return NULL;
@@ -1101,12 +1182,13 @@ static int metadata_property(void *object,
 		if (e == NULL)
 			return -errno;
 
-		e->key = SPA_MEMBER(e, sizeof(*e), void);
+		e->subject = subject;
+		e->key = SPA_PTROFF(e, sizeof(*e), void);
 		strcpy(e->key, key);
-		e->value = SPA_MEMBER(e->key, strlen(e->key) + 1, void);
+		e->value = SPA_PTROFF(e->key, strlen(e->key) + 1, void);
 		strcpy(e->value, value);
 		if (type) {
-			e->type = SPA_MEMBER(e->value, strlen(e->value) + 1, void);
+			e->type = SPA_PTROFF(e->value, strlen(e->value) + 1, void);
 			strcpy(e->type, type);
 		} else {
 			e->type = NULL;
@@ -1138,6 +1220,7 @@ static const struct class metadata_class = {
 	.events = &metadata_events,
 	.destroy = metadata_destroy,
 	.dump = metadata_dump,
+	.name_key = PW_KEY_METADATA_NAME,
 };
 
 static const struct class *classes[] =
@@ -1157,7 +1240,7 @@ static const struct class *find_class(const char *type, uint32_t version)
 {
 	size_t i;
 	for (i = 0; i < SPA_N_ELEMENTS(classes); i++) {
-		if (strcmp(classes[i]->type, type) == 0 &&
+		if (spa_streq(classes[i]->type, type) &&
 		    classes[i]->version <= version)
 			return classes[i];
 	}
@@ -1241,8 +1324,7 @@ static void registry_event_global(void *data, uint32_t id,
 
 bind_failed:
 	pw_log_error("can't bind object for %u %s/%d: %m", id, type, version);
-	if (o->props)
-		pw_properties_free(o->props);
+	pw_properties_free(o->props);
 	free(o);
 	return;
 }
@@ -1264,19 +1346,47 @@ static const struct pw_registry_events registry_events = {
 	.global_remove = registry_event_global_remove,
 };
 
+static bool object_matches(struct object *o, const char *pattern)
+{
+	uint32_t id;
+	const char *str;
+
+	if (spa_atou32(pattern, &id, 0) && o->id == id)
+		return true;
+
+	if (o->props == NULL)
+		return false;
+
+	if (strstr(o->type, pattern) != NULL)
+		return true;
+	if ((str = pw_properties_get(o->props, PW_KEY_OBJECT_PATH)) != NULL &&
+	    fnmatch(pattern, str, FNM_EXTMATCH) == 0)
+		return true;
+	if ((str = pw_properties_get(o->props, PW_KEY_OBJECT_SERIAL)) != NULL &&
+	    spa_streq(pattern, str))
+		return true;
+	if (o->class != NULL && o->class->name_key != NULL &&
+	    (str = pw_properties_get(o->props, o->class->name_key)) != NULL &&
+	    fnmatch(pattern, str, FNM_EXTMATCH) == 0)
+		return true;
+	return false;
+}
+
 static void dump_objects(struct data *d)
 {
-	struct object *o;
-	struct flags_info fl[] = {
+	static const struct flags_info fl[] = {
 		{ "r", PW_PERM_R },
 		{ "w", PW_PERM_W },
 		{ "x", PW_PERM_X },
 		{ "m", PW_PERM_M },
 		{ NULL, },
 	};
+
+	struct object *o;
+
 	d->state = STATE_FIRST;
 	spa_list_for_each(o, &d->object_list, link) {
-		if (d->id != SPA_ID_INVALID && d->id != o->id)
+		if (d->pattern != NULL && !object_matches(o, d->pattern))
 			continue;
 		if (o->changed == 0)
 			continue;
@@ -1355,7 +1465,8 @@ static void show_help(struct data *data, const char *name)
 		"      --version                         Show version\n"
 		"  -r, --remote                          Remote daemon name\n"
 		"  -m, --monitor                         monitor changes\n"
-		"  -N, --no-colors                       disable color output\n",
+		"  -N, --no-colors                       disable color output\n"
+		"  -C, --color[=WHEN]                    whether to enable color support. WHEN is `never`, `always`, or `auto`\n",
 		name);
 }
 
@@ -1371,6 +1482,7 @@ int main(int argc, char *argv[])
 		{ "remote",	required_argument,	NULL, 'r' },
 		{ "monitor",	no_argument,		NULL, 'm' },
 		{ "no-colors",	no_argument,		NULL, 'N' },
+		{ "color",	optional_argument,	NULL, 'C' },
 		{ NULL, 0, NULL, 0}
 	};
 	int c;
@@ -1378,10 +1490,10 @@ int main(int argc, char *argv[])
 	pw_init(&argc, &argv);
 
 	data.out = stdout;
-	if (getenv("NO_COLOR") == NULL)
+	if (isatty(fileno(data.out)) && getenv("NO_COLOR") == NULL)
 		colors = true;
 
-	while ((c = getopt_long(argc, argv, "hVr:mN", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "hVr:mNC", long_options, NULL)) != -1) {
 		switch (c) {
 		case 'h' :
 			show_help(&data, argv[0]);
@@ -1403,15 +1515,27 @@ int main(int argc, char *argv[])
 		case 'N' :
 			colors = false;
 			break;
+		case 'C' :
+			if (optarg == NULL || !strcmp(optarg, "auto"))
+				break; /* nothing to do, tty detection was done
+					  before parsing options */
+			else if (!strcmp(optarg, "never"))
+				colors = false;
+			else if (!strcmp(optarg, "always"))
+				colors = true;
+			else {
+				show_help(&data, argv[0]);
+				return -1;
+			}
+			break;
 		default:
 			show_help(&data, argv[0]);
 			return -1;
 		}
 	}
+
 	if (optind < argc)
-		data.id = atoi(argv[optind++]);
-	else
-		data.id = SPA_ID_INVALID;
+		data.pattern = argv[optind++];
 
 	data.loop = pw_main_loop_new(NULL);
 	if (data.loop == NULL) {
